------------------------------------------------------------------------
-- Copyright Notice:                                                    
-- (c) Copyright 2004, National Optical Astronomy Observatory, which is 
-- operated for the National Science Foundation by the Association of   
-- Universities for Research in Astronomy under Cooperative Agreement   
-- No. AST-0132798.                                                     
--                                                                      
------------------------------------------------------------------------
-- PROJECT      :   MONSOON-TORRENT
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   CFG_SERVICES
-- MODULE       :   DECODE EXECUTION UNIT
-- VERSION      :   2.10
------------------------------------------------------------------------
-- FILE NAME    :   DecodeExecute_V210.vhd
-- AUTHOR         :   PETER MOORE
-- DATE               :   10/02/2011
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.03
------------------------------------------------------------------------
-- DESCRIPTION  :   
-- Deocdes and executes the instruction based on the instruction 
-- definition as specified in the Instruction Set doccument. 
-- The Pattern Pointer which outputs the sequence pattern  
-- is controlled by execution of the data transfer instructions 
-- along with the Clock, Mode, Dev Addr and Board Sel registers.   
--
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
-- Currently the Acknowledge & Request HandShake is commented.
------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	      AUTHOR  VERSION  MODIFICATIONS
-- 01/26/2010   PCM     1.01      Adapted from Orange code version Seq_Mpuv465 to suit Torrent register scheme
--                                                   Modified LDA Opcode behavious to accept a 16-bit operand
--                                                   Modified output format to interface with Wishbone bus multiplexor
--                                                   Modified delay counter constant to provide accurate delays - See note at counter stage !!
-- 10/02/2011  PCM      2.10     Fixed bug that made minimum DSC delay 2 clk cycles.  
------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity DecodeExecute_V210 is
    Port ( 
         SeqCLK          : in  std_logic;                      -- Sequencer code is run synchronous to this clock
         SysClk          : in  std_logic;                      -- Bus clock - sequencer module inputs and outputs synchronous to this clock         
         Reset           : in  std_logic;                      -- Local reset 
         SeqRunFlag      : in  std_logic;                      -- True allows sequencer to run code
         CmdReg          : in  std_logic_vector(15 downto 0);  -- Command register for sequencer         
         OPCODEL         : in  std_logic_vector(15 downto 0);  -- Lower 16-bits of code store output
         OPCODEH         : in  std_logic_vector(15 downto 0);  -- upper 16-bits of code store output
         LoopReg0        : in  std_logic_vector(15 downto 0);  -- one of 16 loop registers for branch control
         LoopReg1        : in  std_logic_vector(15 downto 0);
         LoopReg2        : in  std_logic_vector(15 downto 0);
         LoopReg3        : in  std_logic_vector(15 downto 0);
         LoopReg4        : in  std_logic_vector(15 downto 0);
         LoopReg5        : in  std_logic_vector(15 downto 0);
         LoopReg6        : in  std_logic_vector(15 downto 0);
         LoopReg7        : in  std_logic_vector(15 downto 0);
         LoopReg8        : in  std_logic_vector(15 downto 0);
         LoopReg9        : in  std_logic_vector(15 downto 0);
         LoopReg10       : in  std_logic_vector(15 downto 0);
         LoopReg11       : in  std_logic_vector(15 downto 0);
         LoopReg12       : in  std_logic_vector(15 downto 0);
         LoopReg13       : in  std_logic_vector(15 downto 0);
         LoopReg14       : in  std_logic_vector(15 downto 0);
         LoopReg15       : in  std_logic_vector(15 downto 0);  -- 16 of 16 loop registers for branch control
         PatternData     : in  std_logic_vector(31 downto 0);  -- Data from pattern memory at current pointer address
         SeqEnablesWrite : out std_logic;				            -- WHEN BOARD SELECT = 1, WRITE TO THE ENABLES REGISTER FROM PATTERN MEMORY.
         SeqTestRegWrite : out std_logic;				            -- WHEN BOARD SELECT = 1, WRITE TO THE TESTCONDITION REGISTER FROM PATTERN MEMORY.
         AutoCancelFlag  : in  std_logic;                      -- Used to automatically cancel board selects when set.
         PANADDR         : in  std_logic_vector(9 downto 0);   -- 16-bit address input from PAN for program memory writes / reads
         ProgPtrA        : out std_logic_vector(9 downto 0);   -- Program pointer for lower code store address
         ProgPtrB        : out std_logic_vector(9 downto 0);   -- Program pointer for upper code store address
         PatMemPtr       : inout std_logic_vector(10 downto 0);-- PATTERN POINTER OUTPUT TO THE PATTERN MEMORY          
         PatStackWriteData : out std_logic_vector(10 downto 0);-- Data to stack for pattern pointer stack push operation
         PatStackReadData  : in  std_logic_vector(10 downto 0);-- Data from stack for pattern pointer pop operation
         LASReadData     : in  std_logic_vector(11 downto 0);  -- loop address data from stack stack
         LASWriteData    : out std_logic_vector(11 downto 0);  -- Loop address data to stack
         LSPtr           : inout std_logic_vector(5 downto 0); -- Loop address stack pointer
         LSWrite         : out std_logic;                      -- Write to loop address stack
         LCSReadData     : in  std_logic_vector(15 downto 0);  -- Loop count data from loop stack
         LCSWriteDAta    : out std_logic_vector(15 downto 0);  -- Loop count data to loop stack
         AddrStackPtr    : inout std_logic_vector(5 downto 0); -- Stack pointer for subroutine call stack
         AddrStackReadData  : in  std_logic_vector(11 downto 0);  -- Program pointer data from address stack
         AddrStackWriteData : out std_logic_vector(11 downto 0);  -- Program pointer data to address stack
         AddrStackWrite  : out std_logic;                      -- Write to address stack
         StatusReg       : out std_logic_vector(31 downto 0);  -- sequencer status 
-- Sequencer output product         
         WbWrtDTag_o     : out std_logic_vector(3 downto 0);   -- Module bus source identifier.
         WbWrtAddr_o     : out std_logic_vector(23 downto 0);  -- Module select and address pointer.
         WbWrtData_o     : out std_logic_vector(31 downto 0);  -- Module write data - i.e. sequencer output
         WbWrtReq_o      : out std_logic;                      -- Request for bus write from sequencer
         WbWrtAck_i      : in  std_logic  	                  -- Acknowledge from bus controller for write to wish bone bus
		  );

end DecodeExecute_V210;

architecture Behavioral of DecodeExecute_V210 is

   constant LPP    : std_logic_vector(3  downto 0) := "0001"; 
   constant IPP    : std_logic_vector(3  downto 0) := "0010";
   constant DPP    : std_logic_vector(3  downto 0) := "0011";
   constant LDA    : std_logic_vector(3  downto 0) := "0100";
   constant LMR    : std_logic_vector(3  downto 0) := "0101";
   constant LSR    : std_logic_vector(3  downto 0) := "0110";
                
   constant CAL    : std_logic_vector(3 downto 0)  := "0111";
   constant RET    : std_logic_vector(3 downto 0)  := "1000";
   constant JCB    : std_logic_vector(3 downto 0)  := "1001";
   constant LRB    : std_logic_vector(3 downto 0)  := "1010";
   constant LPB    : std_logic_vector(3 downto 0)  := "1011";
   constant LPE    : std_logic_vector(3 downto 0)  := "1100";
                
   constant DMS    : std_logic_vector(3 downto 0)  := "1101";
   constant DUS    : std_logic_vector(3 downto 0)  := "1111";
   constant DSC    : std_logic_vector(3 downto 0)  := "1110";
   constant NOP    : std_logic_vector(3 downto 0)  := "0000";
                
   constant ZERO10 : std_logic_vector(9 downto 0)  := "0000000000";
   constant ZERO12 : std_logic_vector(11 downto 0) := x"000";
   constant ZERO15 : std_logic_vector(15 downto 1) := "000000000000000";

   constant ONE    : std_logic_vector(11 downto 0) := "000000000001";
   constant TWO    : std_logic_vector(11 downto 0) := "000000000010";
   constant THREE  : std_logic_vector(11 downto 0) := "000000000011";
   constant FOUR   : std_logic_vector(11 downto 0) := "000000000100";
   constant FIVE   : std_logic_vector(11 downto 0) := "000000000101";
   constant SIX    : std_logic_vector(11 downto 0) := "000000000110";
   constant ONE_K  : std_logic_vector(9 downto 0)  := "1111101000";
--   constant ONE_uS : std_logic_vector(7 downto 0)  := "00010100";    -- 20 for a 20MHz clock frequency
--   constant ONE_uS : std_logic_vector(7 downto 0)  := "00110101";    -- 53 for an input clock frequency of 53.125 Mhz
   constant ONE_uS : std_logic_vector(7 downto 0)  := "00011011";    -- 27 for an input clock frequency of 53.125 / 2 Mhz
   
-- CONSTANTS FOR SEQUENCER BUS MODE WORD
   constant SEQ_RST      : std_logic_vector(1 downto 0) := "00";
   constant SEQ_RD       : std_logic_vector(1 downto 0) := "01";
   constant SEQ_WR16     : std_logic_vector(1 downto 0) := "10";
   constant SEQ_WR32     : std_logic_vector(1 downto 0) := "11";
                    
   signal   ZeroClkCtr   : std_logic;
   signal   ZerouSecCtr  : std_logic;
   signal   ZeromsecCtr  : std_logic;
   signal   ZeroSecCtr   : std_logic;
   signal   Run          : std_logic;
          
   signal   OpCode       : std_logic_vector (19 downto 0); -- OPCODE & OPERAND INPUT FROM pipe MODULE           

	signal 	SecCtr	    : std_logic_vector  (11 downto 0);  -- DELAY COUNTER
	signal 	mSecCtr	    : std_logic_vector  (9  downto 0);  -- DELAY COUNTER
	signal 	uSecCtr	    : std_logic_vector  (9  downto 0);  -- DELAY COUNTER
	signal 	ClkCtr	    : std_logic_vector  (7  downto 0);  -- DELAY COUNTER

	signal 	LoopCtr  	 : std_logic_vector  (15 downto 0);  -- LOOP COUNTER
	signal 	LoopReg  	 : std_logic_vector  (15 downto 0);  -- LOOP COUNTER
	signal 	ModeReg  	 : std_logic_vector  ( 1 downto 0);  -- SEQ_MODE REGISTER
	signal 	DevAddrReg   : std_logic_vector  (15 downto 0);  -- SEQ ADDRESS REGISTER
	signal 	SelReg   	 : std_logic_vector  ( 7 downto 0);  -- MODULE SELECT REGISTER

   signal 	ProgCtr  	 : std_logic_vector  (11 downto 0); 
	signal 	ProgCtrONE 	 : std_logic_vector  (11 downto 0); 
	signal 	ProgCtrTWO   : std_logic_vector  (11 downto 0); 
	signal 	ProgCtrTHREE : std_logic_vector  (11 downto 0); 
	signal 	ProgCtrFOUR  : std_logic_vector  (11 downto 0); 
	signal 	ProgCtrFIVE  : std_logic_vector  (11 downto 0); 
    
   signal   JCBFlag      : std_logic;
   signal   DontWait     : std_logic;
   signal   busRequest   : std_logic;
   signal   busReqAck    : std_logic;
   signal   WbWrtReq     : std_logic;
   signal   WrtEnblReq   : std_logic;
   signal   busAddr      : std_logic_vector(23 downto 0);
   signal   busData      : std_logic_vector(31 downto 0);
   signal   WrtEnable16, WrtEnable32, WrtTestReg32 : std_logic;   
begin

----------------------------------------------------------------------------------------------------------------------
-- SEQUENCER OUTPUT STAGE
----------------------------------------------------------------------------------------------------------------------
   WbOutputProc : process(SysClk, Reset)     -- System clock / wishbone synchronous
   begin
      if (Reset = '1') then
         WbWrtReq  <= '0';
         BusReqAck <= '0';
      elsif (rising_edge(SysClk)) then
         BusReqAck <= '0';
         if (WbWrtAck_i = '1') then       -- We are waiting for an acknowledge and it has arrived
            WbWrtReq  <= '0';
            BusReqAck <= '1';             -- pass acknowledge down to the sequencer clock domain
         end if;
         if (WbWrtReq = '0' and busRequest = '1') then
            WbWrtReq  <= '1';             -- Post the request to the bus controller
            WbWrtAddr_o <= busAddr;       -- Capture the address and data fields
            WbWrtData_o <= busData;
         end if;
      end if;
   end process;

   WbWrtDTag_o <= "0000";     -- Anonymous data source tag.
   WbWrtReq_o  <= WbWrtReq;

   SeqOutputProc : process(SeqCLK, Reset, BusReqAck)     -- Sequencer clock / sequencer data synchronous
   begin
      if (Reset = '1' or BusReqAck = '1') then                                     -- cancel bus request asynchronously .. the pulse has a duration of one SysClk
         busRequest <= '0';
      elsif (rising_edge(SeqCLK)) then
         if (SelReg(7 downto 0) = "00000000") then                              -- We have a board select active - go look for a bus cycle
            busRequest <= '0';                                                  -- Cancel any bus request flag if no board is selected.
         else
            case ModeReg is                                                     -- We have a request to write something -  determine which mode we are using and select the address accordingly
               when SEQ_WR16 =>
                  busAddr <= SelReg(7 downto 0) & PatternData(31 downto 16);    -- In 16-bit mode, Combine module select with module address contained in 16-bits most significant pattern word 
                  busData <= ("0000000000000000" & PatternData(15 downto 0));   -- Data is just least significant 16-bits of pattern word
                  busRequest  <= not WrtEnblReq;                                -- Don't bother the wishbone if this is a write to our internal EFR
               when SEQ_WR32 =>
                  busAddr <= SelReg(7 downto 0) & DevAddrReg;                   -- In 32-bit mode, Combine module select with module address contained in LDA regiseter
                  busData <= PatternData;                                       -- with full 32-bit data word from pattern memory data
                  busRequest  <= not WrtEnblReq;
               when others => 
                  busRequest <= '0';                                            -- Do not respond to any reset or read commands .. we cannot use them here.
            end case;
         end if;
      end if;
   end process;
   
   Run <= SeqRunFlag and (not busRequest) and DontWait;  -- Principle sequencer enabling flag.
   
	StatusReg <= "00" & AddrStackPtr(5 downto 0) & "000000" & ProgCtr(9 downto 0) & "00000" & WbWrtReq & busRequest & Run;
   
----------------------------------------------------------------------------------------------------------------------    
-- Write to enables register logic:
----------------------------------------------------------------------------------------------------------------------    
   WrtEnable16 <= '1'  when (PatternData(31 downto 16) = x"0108") else '0';
   WrtEnable32 <= '1'  when (DevAddrReg = x"0108") else '0';
   WrtTestReg32 <= '1' when (DevAddrReg = x"0109") else '0'; 
   WrtEnblReq       <= (WrtEnable16 and SelReg(2)) when (ModeReg = SEQ_WR16) else 
                       (WrtEnable32 and SelReg(2)) when (ModeReg = SEQ_WR32) else '0'; -- Write to the enables register decoded here. - Set to equal CFG_Services module address @ 0x04 + Address 0x0108
	SeqEnablesWrite  <= WrtEnblReq;
   SeqTestRegWrite  <= WrtTestReg32 and SelReg(2);
----------------------------------------------------------------------------------------------------------------------    
-- OPCODE DEMULTIPLEXOR
----------------------------------------------------------------------------------------------------------------------    
    
	process(ProgCtr(1 downto 0), OpCodeL, OpCodeH)
	begin
		case ProgCtr (1 downto 0) is
		  	when "00" =>
			OPCODE(15 downto 0)  <= OpCodeL ;
			OPCODE(19 downto 16) <= OpCodeH(3 downto 0) ;
			when "01" =>
			OPCODE(11 downto 0)  <= OpCodeL(15 downto 4) ;
			OPCODE(19 downto 12) <= OpCodeH(7 downto 0) ;
			when "10" =>
			OPCODE(7 downto 0)   <= OpCodeL(15 downto 8) ;
			OPCODE(19 downto 8)  <= OpCodeH(11 downto 0) ;
			when "11" =>
			OPCODE(3 downto 0)   <= OpCodeL(15 downto 12) ;
			OPCODE(19 downto 4)  <= OpCodeH;
			when others => null;
		end case;
	end process;
    
----------------------------------------------------------------------------------------------------------------------
-- PROGRAM COUNTER MANAGEMENT
----------------------------------------------------------------------------------------------------------------------    
   ProgCtrONE   <= ProgCtr + ONE;
   ProgCtrTWO   <= ProgCtr + TWO;
   ProgCtrTHREE <= ProgCtr + THREE;
   ProgCtrFOUR  <= ProgCtr + FOUR;
   ProgCtrFIVE  <= ProgCtr + FIVE;

   ProgPtrA <= ProgCtr(11 downto 2) when (SeqRunFlag = '1') else PANADDR;
   ProgPtrB <= ProgCtrFOUR(11 downto 2);

------------------------------------
-- PROGRAM POINTER INCREMENT PROCESS
------------------------------------
	process(SeqCLK, RESET)		
   begin
	   if (RESET = '1') then -- Reset is active initialize 
			 ProgCtr <= (others => '0');
      elsif rising_edge(SeqCLK) then
         if (Run = '1') then
       		case OPCODE(3 downto 0) is
    				when IPP | DPP | NOP | LPB =>
    					ProgCtr <= ProgCtrONE; 
    				when LMR | LRB =>
    					ProgCtr <= ProgCtrTWO; 
    				when  DMS | DUS |  DSC |  LSR =>
    					ProgCtr<= ProgCtrTHREE; 
               when  LPP => 
    					ProgCtr <= ProgCtrFOUR;
               when LDA =>
    					ProgCtr <= ProgCtrFIVE;               
    				when CAL => 
    					ProgCtr <= OpCode(15 downto 4);
    				when RET =>
    					ProgCtr <= AddrStackReadData;
    				when JCB =>
    					if (JCBFlag = '1') then
    						ProgCtr <= OpCode(15 downto 4);  
    					else 
    						ProgCtr <= ProgCtrFIVE;  
    					end if;
    				when LPE => 
    					if (LoopCtr(15 downto 1) = ZERO15) then
    						ProgCtr <= ProgCtrONE;
    					else
    						ProgCtr <= LASReadData; 
    					end if;
    				when others => null;
    			end case;
			elsif (SeqRunFlag = '0') then
				ProgCtr <= (others => '0');                        -- Initialize program counter to zero when not running.
         end if;
      end if;
	end process; 

--------------------------------
-- PROGRAM POINTER STACK PROCESS
--------------------------------
	process(RESET, SeqCLK)		
    begin
		if( RESET = '1' ) then -- Reset is active initialize 
            AddrStackPtr <= (others => '0');
			AddrStackWriteData <= (others => '0');
			AddrStackWrite <= '0';
		elsif rising_edge(SeqCLK) then
			AddrStackWrite <= '0';
           if (Run = '1') then
       			case OPCODE ( 3 downto 0)  is
             		when CAL=> 
						AddrStackPtr <= AddrStackPtr + '1';
						AddrStackWriteData <= ProgCtrFOUR;
					    AddrStackWrite <= '1';
					when RET =>
					    AddrStackPtr <= AddrStackPtr - '1';
                    when others => null;
    			end case;
            elsif (SeqRunFlag = '0') then
                AddrStackPtr <= (others => '0');
			    AddrStackWrite <= '0';
    		end if;
        end if;
	end process; 

----------------------------------------------------------------------------------------------------------------------
-- PATTERN POINTER MANAGEMENT
----------------------------------------------------------------------------------------------------------------------
	process(RESET, SeqCLK)		
    begin
		if (RESET = '1') then -- Reset is active initialize 
			PatMemPtr <= (others => '0');
		elsif rising_edge(SeqCLK) then
            if (Run = '1') then
       			case OPCODE (3 downto 0)  is
    				when IPP =>
    					PatMemPtr <= PatMemPtr + 1; 
    				when DPP =>
    					PatMemPtr <= PatMemPtr - 1; 
    				when LPP =>
    					PatMemPtr <= OpCode(14 downto 4); 
             		when CAL=> 
						PatStackWriteData <= PatMemPtr;
					when RET =>
					    PatMemPtr <= PatStackReadData;
                    when  others => null;
    			end case;
            elsif (SeqRunFlag = '0') then
			    PatMemPtr <= (others => '0');                       -- Initialize pattern pointer to 0 when not running.
    		end if;
        end if;
	end process; 
    
----------------------------------------------------------------------------------------------------------------------
-- SEQ BUS MODE WORD MANAGEMENT 
----------------------------------------------------------------------------------------------------------------------
	process(RESET, SeqCLK)		
   begin
      if (RESET = '1') then -- Reset is active initialize 
			ModeReg <= SEQ_WR32;                          -- Initialize Mode register to WRT32 command
		elsif rising_edge(SeqCLK) then
         if (Run = '1') then
            case OPCODE ( 3 downto 0)  is
               when LMR =>
                  ModeReg <= OPCODE (5 downto 4);
               when others => 
                  null;
				end case;
         elsif (SeqRunFlag = '0') then
            ModeReg <= SEQ_WR32;                         -- Initialize Mode register to WRT32 command
			end if;
     end if;
  end process; 

----------------------------------------------------------------------------------------------------------------------
-- SEQ BUS DEVICE ADDRESS MANAGMENT
----------------------------------------------------------------------------------------------------------------------
	process(RESET, SeqCLK)		
    begin
		if (RESET = '1') then -- Reset is active initialize 
			DevAddrReg <= (others => '0');
		elsif rising_edge(SeqCLK) then
			if (Run = '1') then
            case OPCODE (3 downto 0)  is
					when LDA=>
	               DevAddrReg <= OPCODE (19 downto 4);
	            when others => 
	               null;
				end case;
         elsif (SeqRunFlag = '0') then
            DevAddrReg <= (others => '0');                      -- Initialize Device Addr word to zero.
			end if;
      end if;
	end process; 

---------------------------------------------------------------------------------------------------------------------- 
-- SEQ BUS BOARD SELECT MANAGEMENT
----------------------------------------------------------------------------------------------------------------------
	process(RESET, SeqCLK)		
    begin
		if (RESET = '1') then -- Reset is active initialize 
			SelReg <= (others => '0');
		elsif rising_edge(SeqCLK) then
			if (Run = '1') then
   			case OPCODE(3 downto 0)  is
               when LSR=>
	               SelReg <= OPCODE(11 downto 4);
	             when others =>
                  if (AutoCancelFlag = '1') then
             			SelReg <= (others => '0');
                  else
	                  SelReg <= SelReg;
                  end if;
				end case;
         elsif (SeqRunFlag = '0') then
			   SelReg <= (others => '0');
         end if;
      end if;
	end process; 

---------------------------------------------------------------------------------------------------------------------- 
-- CONDITIONAL JUMP FLAG MULTIPLEXOR
----------------------------------------------------------------------------------------------------------------------
    process(CmdReg, OpCode ( 19 downto 16))
    begin
       case OpCode ( 19 downto 16) is								
		   when "0000" => JCBFlag <= CmdReg(0);
         when "0001" => JCBFlag <= CmdReg(1);
		   when "0010" => JCBFlag <= CmdReg(2);
		   when "0011" => JCBFlag <= CmdReg(3);
		   when "0100" => JCBFlag <= CmdReg(4);
		   when "0101" => JCBFlag <= CmdReg(5);
		   when "0110" => JCBFlag <= CmdReg(6);
		   when "0111" => JCBFlag <= CmdReg(7);
		   when "1000" => JCBFlag <= CmdReg(8);
         when "1001" => JCBFlag <= CmdReg(9);
		   when "1010" => JCBFlag <= CmdReg(10);
		   when "1011" => JCBFlag <= CmdReg(11);
		   when "1100" => JCBFlag <= CmdReg(12);
		   when "1101" => JCBFlag <= CmdReg(13);
		   when "1110" => JCBFlag <= CmdReg(14);
		   when "1111" => JCBFlag <= CmdReg(15);		
		   when others => null;					
		end case;					
    end process;
      
----------------------------------------------------------------------------------------------------------------------
-- LOOP COUNTER AND REGISTER MANAGEMENT
----------------------------------------------------------------------------------------------------------------------
-----------------------	
-- COUNTER LOAD PROCESS
-----------------------
	process(RESET, SeqCLK)		
    begin
		if( RESET = '1' ) then -- Reset is active initialize 
            LoopCtr <= (others => '0');
    	elsif rising_edge(SeqCLK) then
            if (Run = '1') then
                case OPCODE(3 downto 0)  is
                    when LRB =>
                    	LoopCtr <= LoopReg;                      -- 16 bit loop counter register contents	
                    when LPB =>
                    	LoopCtr <= PatternData(15 downto 0);     -- 16 bit loop count from pattern memory data
                    when LPE => 
                    	if (LoopCtr(15 downto 1) = ZERO15) then
                    		LoopCtr <= LCSReadData;
                    	else
							LoopCtr <= LoopCtr - 1;
                    	end if;
                    when others => null;
                end case;
            elsif (SeqRunFlag = '0') then
                LoopCtr <= (others => '0');
            end if;
        end if;                
    end process;
        
------------------------------------------------   	
-- LOOP COUNTER STACK POINTER MANAGEMENT PROCESS
------------------------------------------------
	process(RESET, SeqCLK)		
   begin
		if (Reset = '1') then
			LSPtr   <= (others => '0');
			LSWrite <= '0';
			LASWriteData <= (others => '0');
			LCSWriteData <= (others => '0');
      elsif rising_edge(SeqCLK) then
         LSWrite <= '0';
         if (Run = '1') then
            case OPCODE(3 downto 0)  is
               when LRB =>
               	LSWrite <= '1';
               	LSPtr   <=  LSPtr + 1;
               	LASWriteData <= ProgCtrTWO;					   
               	LCSWriteData <= LoopCtr;
               when LPB =>
               	LSWrite <= '1';
               	LSPtr   <=  LSPtr + 1;
               	LASWriteData <= ProgCtrONE;						   
               	LCSWriteData <= LoopCtr;
               when LPE => 
               	if (LoopCtr(15 downto 1) = ZERO15) then
                  	LSPtr <= LSPtr - 1;
                  end if;
               when others => null;
            end case;
         elsif (SeqRunFlag = '0') then
    			LSPtr   <= (others => '0');
    			LSWrite <= '0';
         end if;
      end if;                
   end process;

--------------------------------------
-- LOOP COUNTER REGISTER DEMULTIPLEXOR
--------------------------------------
   process(LoopReg0, LoopReg1, LoopReg2, LoopReg3, LoopReg4, LoopReg5, LoopReg6, LoopReg7, LoopReg8,
    LoopReg9, LoopReg10, LoopReg11, LoopReg12, LoopReg13, LoopReg14, LoopReg15, OPCODE(7 downto 4))
   begin
      case OPCODE(7 downto 4) is								
		   when "0000" => LoopReg <= LoopReg0;
         when "0001" => LoopReg <= LoopReg1;
		   when "0010" => LoopReg <= LoopReg2;
		   when "0011" => LoopReg <= LoopReg3;
		   when "0100" => LoopReg <= LoopReg4;
		   when "0101" => LoopReg <= LoopReg5;
		   when "0110" => LoopReg <= LoopReg6;
		   when "0111" => LoopReg <= LoopReg7;
		   when "1000" => LoopReg <= LoopReg8;
         when "1001" => LoopReg <= LoopReg9;
		   when "1010" => LoopReg <= LoopReg10;
		   when "1011" => LoopReg <= LoopReg11;
		   when "1100" => LoopReg <= LoopReg12;
		   when "1101" => LoopReg <= LoopReg13;
		   when "1110" => LoopReg <= LoopReg14;
		   when "1111" => LoopReg <= LoopReg15;		
		   when others => null;					
		end case;					
    end process;

----------------------------------------------------------------------------------------------------------------------
--DELAY COUNTERS
----------------------------------------------------------------------------------------------------------------------
	process(RESET, SeqRunFlag, SeqClk)		
   begin
      if (RESET = '1' or SeqRunFlag = '0') then -- Reset is active initialize 
      	 ClkCtr  <= (others => '0');
          uSecCtr <= (others => '0');
          msecCtr <= (others => '0');
          SecCtr <= (others => '0');
      elsif rising_edge (SeqCLK) then
         if (Run = '1') then
            case OPCODE(3 downto 0) is
            	when DMS =>
            		mSecCtr(7 downto 0) <= Opcode(11 downto 4);
            	when DUS =>
            	   uSecCtr(7 downto 0) <= Opcode(11 downto 4);
            	when DSC =>
--                  if (Opcode(11 downto 4) > "00000001") then  -- Corrects problem with the minimum DSC period being 2 clock cycles.
                  ClkCtr(7 downto 0)  <= Opcode(11 downto 4);
--                  else
--                     ClkCtr  <= (others => '0');
--                  end if;
               when others => null;		
            end case;
         end if;
      	if (ZeroClkCtr = '0') then  
            ClkCtr <= ClkCtr - 1;         -- Note. For 53.125 MHz clock frequency we loose one clock period of accuracy every eight clock cycles (i.e. 28ns). Might want to fix this later :-)    
  			elsif (ZerouSecCtr = '0') then
            uSecCtr <= uSecCtr - 1;
    			ClkCtr  <= ONE_uS;	
  			elsif (ZeromsecCtr = '0') then
			   mSecCtr <= mSecCtr - 1;
   			uSecCtr <= ONE_K;
   			ClkCtr  <= ONE_uS;			
  			elsif (ZeroSecCtr = '0') then	
            SecCtr <= SecCtr - 1;
				mSecCtr <= ONE_K;
				uSecCtr <= ONE_K;
				ClkCtr  <= ONE_uS;			
			end if;
      end if;
   end process; 

   ZeroClkCtr  <= '1' when (ClkCtr = ZERO10) else '0';
   ZerouSecCtr <= '1' when (uSecCtr = ZERO10) else '0';
   ZeromsecCtr <= '1' when (mSecCtr = ZERO10) else '0';
   ZeroSecCtr  <= '1' when (SecCtr = ZERO12) else '0';
   DontWait <= ZeroClkCtr and ZerouSecCtr and ZeromsecCtr and ZeroSecCtr;  

end Behavioral;
