------------------------------------------------------------------------
-- Copyright Notice:                                                    
-- (c) Copyright 2004, National Optical Astronomy Observatory, which is 
-- operated for the National Science Foundation by the Association of   
-- Universities for Research in Astronomy under Cooperative Agreement   
-- No. AST-0132798.                                                     
--                                                                      
------------------------------------------------------------------------
-- PROJECT      :   MONSOON-TORRENT
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   CFG_SERVICES
-- MODULE       :   ENABLES REGISTER
-- VERSION      :   2.11
------------------------------------------------------------------------
-- FILE NAME    :   EnablesReg_Ver211.VHD
-- AUTHOR         :   PETER MOORE
-- DATE               :   02/27/2012
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.03
------------------------------------------------------------------------
-- DESCRIPTION  :   
-- THIS REGISTER CONTAINS SOME STROBES AND CONTROL BITS USED IN CONTROL 
-- CONTEXT FOR CERTAIN SEQ_MPU FUNCTIONS. THE WRITE TO THIS REGISTER IS
-- FROM THE SEQ_MPU ITSELF WHEN THE BOARD SELECT VALUE = 1 (I.E. MCB).
-- DATA IS LOADED FROM THE CURRENT PATTERN POINTER AND PROCESSED ACCORDING
-- TO ITS' BIT POSITION AS OUTLINED BELOW.
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
-- PATTERN MEMORY BITS AND THEIR FUNCTION
-- BIT								BIT
-- 31								15 FORCE SYNC OUT HIGH
-- 30								14 FORCE SYNC OUT LOW
-- 29								13 DISABLE PIXEL DATA STREAM
-- 28								12 ENABLE PIXEL DATA STREAM 
-- 27								11  RESET READOUT  IN PROGRESS FLAG
-- 26								10 SET READOUT  IN PROGRESS FLAG
-- 25								 9  PULSE LINE START STROBE
-- 24								 8  PULSE FRAME START STROBE
-- 23								 7  SET PREFLASH OFF 
-- 22								 6  SET PREFLASH ON
-- 21								 5  CLOSE SHUTTER
-- 20								 4  OPEN SHUTTER 
-- 19								 3  STOP INTEGRATION TIMER 
-- 18								 2  START INTEGRATION TIMER 
-- 17								 1   PULSE POWER SUPPLY SYNC LOGIC
-- 16								 0  RESET START EXP FLAG IN CMD REG
------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	        AUTHOR  VERSION  MODIFICATIONS
-- 01/26/2010   PCM     1.01      Adapted from Orange code version Seq_Mpuv465 to suit Torrent register scheme.
--                                                  Added shutter and preflash control bits.
-- 03/05/2010   PCM     1.03     Changed bit order, eliminated the LSR_AutoKill, added FrameStart and LineStart features.
--  10/02/2011   pcm      2.10    Updated the IMPLEMENTATION NOTES to make them true ! Added the pixel data disable flag
--  02/27/2012  PCM     2.11    Added pause signal to disable the shutter and preflash during an integration pause. 
--
------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity EnablesReg_Ver211 is
    Port (SeqClk_i	 		: in  std_logic;
         SeqRunFlag_i      : in  std_logic;  -- True when sequencer is actually running code
         WrtEnable_i 	   : in  std_logic;  -- STROBE TRUE WHEN BOARD SELECT = 1 AND MODE = WRT32
         PatternData_i     : in  std_logic_vector(31 downto 0);  -- Input code to select enables register action
         EnableIntCtr_o 	: out std_logic;  -- ENABLE LINE TO INTEGRATION TIMER, STARTS AND STOPS INTEGRATION TIMER
         AutoCancel_o      : out std_logic;  -- Set the flag to automatically cancel board select strobes after one clk cycle
         ResetSEBit_o      : out std_logic;  -- STROBE TO RESET THE START EXPOSURE FLAG IN CMD REGISTER 
         DHESyncOut_o 	   : out std_logic;  -- STROBE TO GO TO SYNC LINE ON OTHER DHEs'
			PSMSyncOut_o      : out std_logic;  -- Strobe to maintain synchronization to the power supply switching frequency.
         OpenShutter_o     : out std_logic;  -- Command to open the shutter
         SetPreflashOn_o   : out std_logic;  -- Command to set preflash control bit on.
         ReadoutBusy_o     : out std_logic;  -- Bit to inform the system that we are busy reading out - suspend all 'noisey' processes etc.
         EnablesReg_o      : out std_logic_vector(7 downto 0); -- Copy of the enables register flags for monitoring by the PAN
         FrameSync_o       : out std_logic;  -- Indicates a new frame beginning
         LineSync_o        : out std_logic;  -- Indicates a new row beginning
         PixDataDisable_o  : out std_logic;   -- Disables the pixel data stream to the pan
         Pause_i           : in  std_logic
         );
end EnablesReg_Ver211;

architecture Behavioral of EnablesReg_Ver211 is
   constant SYNC_PERIOD : std_logic_vector(5 downto 0) := "110101";  -- approx. 2us Frame and Line sync period
   signal EnReg : std_logic_vector(7 downto 0);
   signal FrameSync, LineSync : std_logic;
   signal FrameCntr, LineCntr : std_logic_vector(5 downto 0);
begin

-- Enables register assignments
   ResetSEBit_o     <= EnReg(0); 
   EnableIntCtr_o   <= EnReg(1);
   OpenShutter_o    <= EnReg(2) and not Pause_i; 
   SetPreflashOn_o  <= EnReg(3) and not Pause_i;
   AutoCancel_o     <= '1'; 
   ReadoutBusy_o    <= EnReg(5);  
   PSMSyncOut_o     <= EnReg(6); 
   DHESyncOut_o     <= EnReg(7);
   EnablesReg_o     <= EnReg;
   FrameSync_o      <= FrameSync;
   LineSync_o       <= LineSync;
   PixDataDisable_o <= EnReg(4);
   
------------------------------------------------------------------------
-- RESET THE START EXPOSURE FLAG IN THE CMD REGISTER WHEN BIT 0 HIGH
------------------------------------------------------------------------
	start_exp : process (SeqClk_i)
	begin
		if (rising_edge(SeqClk_i)) then
			if (WrtEnable_i = '1') then
				EnReg(0) <= PatternData_i(0);
			else
				EnReg(0) <= '0';
			end if;
		end if;
	end process;

------------------------------------------------------------------------
-- PULSE THE POWER SUPPLY SYNC FLAG WHEN BIT 1 HIGH
------------------------------------------------------------------------
	pwr_sync_process : process (SeqClk_i)
	begin
		if (rising_edge(SeqClk_i)) then
			if (WrtEnable_i = '1') then
				EnReg(6) <= PatternData_i(1);
			else
				EnReg(6) <= '0';
			end if;
		end if;
	end process;
   
------------------------------------------------------------------------
-- ENABLE THE INTEGRATION TIMER COUNTER WHEN BIT 2 HIGH
-- DISABLE THE INTEGRATION TIMER COUNTER WHEN BIT 3 HIGH - HAS PRIORITY
------------------------------------------------------------------------
	int_rs : process (SeqClk_i)
	begin
		if (rising_edge(SeqClk_i)) then
         if (SeqRunFlag_i = '1') then
    			if (WrtEnable_i = '1' and PatternData_i(3) = '1') then
    				EnReg(1) <= '0';
            elsif (WrtEnable_i = '1' and PatternData_i(2) = '1') then
    				EnReg(1) <= '1';
    			end if;
         else
            EnReg(1) <= '0';
         end if;    
		end if;
	end process;
   
------------------------------------------------------------------------
-- SET THE SHUTTER OPEN BIT HIGH WHEN BIT 4 HIGH - HAS PRIORITY
-- SET THE SHUTTER OPEN BIT LOW WHEN BIT 5 HIGH
------------------------------------------------------------------------
	shutter_rs : process (SeqClk_i)
	begin
		if (rising_edge(SeqClk_i)) then
         if (SeqRunFlag_i = '1') then
    			if (WrtEnable_i = '1' and PatternData_i(4) = '1') then
    				EnReg(2) <= '1';
    			elsif (WrtEnable_i = '1' and PatternData_i(5) = '1') then
    				EnReg(2) <= '0';
    			end if;
         else
    		   EnReg(2) <= '0';
         end if;
		end if;
	end process;
   
------------------------------------------------------------------------
-- SET THE PREFLASH ON BIT HIGH WHEN BIT 6 HIGH - HAS PRIORITY
-- SET THE PREFLASH ON BIT LOW WHEN BIT 7 HIGH
------------------------------------------------------------------------
	preflash_rs : process (SeqClk_i)
	begin
		if (rising_edge(SeqClk_i)) then
         if (SeqRunFlag_i = '1') then
    			if (WrtEnable_i = '1' and PatternData_i(6) = '1') then
    				EnReg(3) <= '1';
    			elsif (WrtEnable_i = '1' and PatternData_i(7) = '1') then
    				EnReg(3) <= '0';
    			end if;
         else
    		   EnReg(3) <= '0';
         end if;
		end if;
	end process;
  
------------------------------------------------------------------------
-- SEND A FRAME START FLAG SIGNAL WHEN BIT 8 HIGH
------------------------------------------------------------------------   
	Frame_sync_process : process (SeqClk_i)
	begin
		if (rising_edge(SeqClk_i)) then
         if (FrameCntr < SYNC_PERIOD) then
            FrameSync <= '1';
            FrameCntr <= FrameCntr + 1;
			elsif (WrtEnable_i = '1' and PatternData_i(8) = '1') then
				FrameCntr <= (others => '0');
			else
				FrameSync <= '0';
			end if;
		end if;
	end process;
   
-------------------------------------------------------------------------
-- SEND A LINE START FLAG SIGNAL WHEN BIT 9 HIGH
------------------------------------------------------------------------   
	Line_sync_process : process (SeqClk_i)
	begin
		if (rising_edge(SeqClk_i)) then
         if (LineCntr < SYNC_PERIOD) then
            LineSync <= '1';
            LineCntr <= LineCntr + 1;
			elsif (WrtEnable_i = '1' and PatternData_i(9) = '1') then
				LineCntr <= (others => '0');
			else
				LineSync <= '0';
			end if;
		end if;
	end process;

------------------------------------------------------------------------
-- SET THE READOUT BUSY FLAG WHEN BIT 10 HIGH - HAS PRIORITY
-- RESET THE READOUT BUSY FLAG WHEN BIT 11 HIGH
------------------------------------------------------------------------
	readBusy_rs : process (SeqClk_i)
	begin
		if (rising_edge(SeqClk_i)) then
         if (SeqRunFlag_i = '1') then
    			if (WrtEnable_i = '1' and PatternData_i(10) = '1') then
    				EnReg(5) <= '1';
    			elsif (WrtEnable_i = '1' and PatternData_i(11) = '1') then
    				EnReg(5) <= '0';
            end if;
         else
    		   EnReg(5) <= '0';
         end if;
		end if;
	end process;

------------------------------------------------------------------------
-- SET THE PIXEL DATA DISABLE FLAG  HIGH WHEN BIT 13 HIGH
-- SET THE PIXEL DATA DISABLE  LOW WHEN BIT 12 HIGH - HAS PRIORITY
------------------------------------------------------------------------
	pixDataDisable_rs : process (SeqClk_i)
	begin
		if (rising_edge(SeqClk_i)) then
         if (SeqRunFlag_i = '1') then
    			if (WrtEnable_i = '1' and PatternData_i(12) = '1') then
    				EnReg(4) <= '0';
    			elsif (WrtEnable_i = '1' and PatternData_i(13) = '1') then
    				EnReg(4) <= '1';
    			end if;
         else
    		    EnReg(4) <= '0';
         end if;
		end if;
	end process;
   
------------------------------------------------------------------------
-- SET THE SYNC_OUT FLAG  HIGH WHEN BIT 15 HIGH
-- SET THE SYNC_OUT FLAG  LOW WHEN BIT 14 HIGH - HAS PRIORITY
------------------------------------------------------------------------
	sync_rs : process (SeqClk_i)
	begin
		if (rising_edge(SeqClk_i)) then
         if (SeqRunFlag_i = '1') then
    			if (WrtEnable_i = '1' and PatternData_i(14) = '1') then
    				EnReg(7) <= '0';
    			elsif (WrtEnable_i = '1' and PatternData_i(15) = '1') then
    				EnReg(7) <= '1';
    			end if;
         else
    		    EnReg(7) <= '0';
         end if;
		end if;
	end process;
   
end Behavioral;
