------------------------------------------------------------------------
-- Copyright Notice:
-- (c) Copyright 2004 - 2008, National Optical Astronomy Observatory,
-- which is operated for the National Science Foundation by the
-- Association of Universities for Research in Astronomy under
-- Cooperative Agreement No. AST-0132798.
------------------------------------------------------------------------
-- PROJECT      :   MONSOON-TORRENT
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   CFG_SERVICES
-- MODULE       :   I2C INTERFACE
-- VERSION      :   2.10
------------------------------------------------------------------------
-- FILE NAME    :   I2C_Interface_V210.vhd
-- AUTHOR         :   Peter Moore
-- DATE               :   10/18/2011
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.03
------------------------------------------------------------------------
-- DESCRIPTION  :  This is the serializer / deserializer that transmits or receives I2C bit streams
--  Addapted from orginal MONSOON Orange code - SerConfigIfcV42                    
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
--
------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	     AUTHOR  VERSION  MODIFICATIONS
-- 06/26/2009  PCM     1.00    INITIAL VERSION FOR TORRENT ADAPTED FROM SERCONFIGIFC.VHD VERSION 4.2
-- 10/18/2011    PCM      2.10   Cleaned up the RdLoadFlag register which was never used.
------------------------------------------------------------------------
-- PORT DEFINITIONS :
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity I2C_Interface_V210 is
	Port (
      SysClk_i     : in  std_logic;
      Reset_i      : in  std_logic;
-- Module Control
      ChannelAddr_i: in  std_logic_vector(4 downto 0);  -- Bit select address of module to send / receive data -- note that more than one port can be selected
      DeviceAddr_i : in  std_logic_vector(6 downto 0);   -- Peripheral bus address
      GoRead_i     : in  std_logic; -- Initiate a read operation from the addressed bus address.
      GoWrite_i    : in  std_logic; -- Initiate a write operation to the addressed bus address
      RdWrtBytes_i : in  std_logic_vector(6 downto 0);  -- Number of bytes to read or write to the bus device  (i.e. max 127 bytes per transaction).
-- Read Data interface
      DataOut_o    : out std_logic_vector(39 downto 0); -- Data read from the bus device (8 bits x 5 ports)
      DataOutRdy_o : out std_logic;    -- Data is valid on the DataOut signals
      DataOutAck_i : in  std_logic;    -- Acknowledge from controller upon accepting the byte
-- Write data interface
      DataIn_i     : in  std_logic_vector(7 downto 0); -- Data to write to the bus device      
      DataInRdy_i  : in  std_logic;    -- Write data is valid on the DataIn signals
      DataInAck_o  : out std_logic;    -- Acknowledge from this interface upon accepting the byte
-- I2C interface status signals
      IfcBusy_o    : out std_logic;
      IfcRdErr_o   : out std_logic_vector(4 downto 0); -- An error occured while reading the bus device (no good acknowledge)
      IfcWrtErr_o  : out std_logic_vector(4 downto 0); -- An error occured while writing to the bus device (no good acknowledge)
-- I2C interface
      MstrActive_o : out std_logic;
      MstrWrtEn_o  : out std_logic;  -- enable SDA_o onto the peripheral bus
      MstrScl_o    : out std_logic;  -- I2C master clock
      MstrSda_i    : in  std_logic_vector(4 downto 0);  -- I2C master data input
      MstrSda_o    : out std_logic;   -- I2C master data output
      MuxBusy_i    : in  std_logic    -- Input from mux indicating busy status      
   );
end I2C_Interface_V210;

architecture Behavioral of I2C_Interface_V210 is

-- Constants used in the logic here
   constant CLK_PERIOD    : std_logic_vector(7 downto 0) := "10100000";  -- Serial clock phase period in terms of sysclk. Nominally 160 (1000 ns).
   constant RD_STROBE     : std_logic_vector(7 downto 0) := "01010000";  -- When device data is readable.
   constant EEPROM        : std_logic_vector(6 downto 0) := "1010100";   -- Fixed device bus address for eeprom devices - Controls the time between write operations.
   constant BITS_PER_BYTE : std_logic_vector(3 downto 0) := "1000";      -- Number of bits in a byte ... Really !
-- Control registers and signals
   signal SerCfgDev   : std_logic_vector(6 downto 0); -- Current device address in use.
   signal BusPort, DataValidPort : std_logic_vector(4 downto 0);
   signal IfcBusy     : std_logic;
   signal CmdEnable   : std_logic;
   
-- Write eeprom FSM signals
   type CfgWrtStateType is (CfgWrtIdle, CfgWrtWordLoop, CfgWrtDataByte, CfgWrtDataWait, CfgWrtSendStop, CfgWrtDelay);
   signal CfgWrtState  : CfgWrtStateType;
   signal BUSFREE_PERIOD : std_logic_vector(19 downto 0);   -- period of time between bus transactions = 5.1ms for eeprom writes, 1.5us for others.
   signal SerCfgWrt    : std_logic;
   signal WrtRomActive : std_logic;                        -- Indicates the write eeprom FSM is active.
   signal WrtLoadData  : std_logic_vector (7 downto 0);
   signal WrtLoadFlag  : std_logic;
   signal WrtClkEnable : std_logic;
   signal CfgWrtAddr   : std_logic_vector(6 downto 0);
   signal WrtStartReq  : std_logic;                        -- Trigger a start + select sequence
   signal WrtStopReq   : std_logic;
   signal WrtDlyCntr   : std_logic_vector(19 downto 0);    -- Used to time out bus free periods.
-- Read eeprom FSM signals
   type CfgRdStateType is(CfgRdIdle, CfgRdStart, CfgRdRdDat1, CfgRdRdDat2, CfgRdSendStop, CfgRdDelay);
   signal CfgRdState   : CfgRdStateType;
   signal SerCfgRd     : std_logic;
   signal CfgRdAddr    : std_logic_vector(6 downto 0);
--   signal CfgReadStatus: std_logic_vector(15 downto 0);
   signal RdRomActive  : std_logic;
   signal RdClkEnable  : std_logic;
   signal RdStartReq   : std_logic;
   signal RdStopReq    : std_logic;
   signal AckFlag      : std_logic;
--   signal RdSndStartWriteFlag : std_logic;
--   signal FirstFlag    : std_logic;
   signal RdDlyCntr    : std_logic_vector(19 downto 0);    -- Used to time out bus free periods.
-- Start sequence process signals
   type StrtStateType is (StrtIdle, StrtSync, StrtStart, StrtSendSlct);
   signal StrtState : StrtStateType;
   signal SendStartReq : std_logic;         -- Global request for a start sequence
   signal StartBusy    : std_logic;         -- signal that stop sequence is in progress
   signal StrtLoadFlag : std_logic;         -- Request from stop sequence to load data to sda line
   signal StrtClkEnable: std_logic;         -- Request from stop generator for use of clock generator
   signal StrtLoadData : std_logic_vector (7 downto 0);  -- Data for stop generator
   signal WriteFlag    : std_logic;
-- Stop sequence process signals
   type SndStopStateType is (SndStopIdle, SndStopStart, SndStopStop);
   signal SndStopState: SndStopStateType;
   signal SendStopReq : std_logic;         -- Global request for a stop sequence
   signal StopBusy    : std_logic;         -- signal that stop sequence is in progress
   signal StpLoadFlag : std_logic;         -- Request from stop sequence to load data to sda line
   signal StpClkEnable: std_logic;         -- Request from stop generator for use of clock generator
   signal StpLoadData : std_logic_vector (7 downto 0);  -- Data for stop generator
-- SerCfgClkGen process signals
   signal ClkEnableFlag: std_logic;                    -- Enables the clock generator to run.
   signal ClkCntr      : std_logic_vector(7 downto 0); -- Counts sysclk periods to derive the I2C serial clock.
   signal SerCfgClk    : std_logic_vector(1 downto 0); -- Generates three clock phases .. 00, 01, 10 Bit 0 is inverted and used as the SCL clock. 
   signal RdStrobe     : std_logic;                    -- Ocurres in middle of clock phase 01 when SDA data should be read.
   signal ClkPhase0, ClkPhase1, ClkPhase2 : std_logic; -- Decoded clock phase information.
-- SerCfgWrtData process signals
   type WrtDataStateType is (WrtDataIdle, WrtDataBusy, WrtDataAcknowledge, WrtDataEndFrame);
   signal WrtDataState : WrtDataStateType;
   signal WrtSdaEnable : std_logic;
   signal WriteBusy    : std_logic;
   signal LoadData     : std_logic_vector(7 downto 0);
   signal LoadFlag     : std_logic;                    -- Global load flag for write process
   signal SerWrtData   : std_logic_vector(7 downto 0); -- Parallel to serial conversion register.
   signal DataArm      : std_logic;                    -- One shot toggle to slow things down
   signal BitCntr      : std_logic_vector(3 downto 0); -- Serial SDA bit counter
   signal SdaEnable    : std_logic;                    -- Enables control over the sda data line
   signal SdaAckBit    : std_logic;                    -- muxed channel address bits to form one unified acknowledge bit
   signal SdaAck       : std_logic;                    -- Result of acknowledge bit time from slave device after latching at appropriate time.
-- SerCfgRdData process signals
   signal SerRdData_0, SerRdData_1, SerRdData_2, SerRdData_3, SerRdData_4   : std_logic_vector(7 downto 0);  -- Actual serial to parallel conversion register.
   signal RdByteValue_0, RdByteValue_1, RdByteValue_2, RdByteValue_3, RdByteValue_4 : std_logic_vector(7 downto 0);  -- Latched value of byte read in by process. Valid with RdDataRdy.
   signal RdBitCntr   : std_logic_vector(3 downto 0);  -- Counts the number of bits read in.
   signal RdEnableFlag: std_logic;                     -- Lets the read process run and collect data
   signal RdDataRdy   : std_logic;                     -- Flag to indicate that data is ready to be played with
   signal ShiftFlag   : std_logic;                     -- Used to control serial > parallel shift register.
   signal AckSdaEnable: std_logic;                     -- Enables the tristate buffer for SDA master ACKNOWLEDGE signal
   signal RdAckActive : std_logic;                     -- Used during the read acknowledge cycle.

begin
   
------------------------------------------------------------------------
-- PAN Initiated read and write process triggers and
-- EEPROM physical device multiplexor.
------------------------------------------------------------------------

   IfcBusy <= RdRomActive or WrtRomActive or MuxBusy_i;
   IfcBusy_o <= IfcBusy;
   CmdEnable <= not (RdRomActive or WrtRomActive or MuxBusy_i or Reset_i);
   MstrActive_o <= RdRomActive or WrtRomActive;
   
   SerCfgTrigRd : process (SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (CmdEnable = '1' and ChannelAddr_i /= "00000") then  -- We're currently in idle state and have a valid channel address
            if (GoRead_i = '1') then
               SerCfgDev <= DeviceAddr_i;   -- Catch the relevent physical device address.
               BusPort   <= ChannelAddr_i;  -- and the bus port that we are interested in
               SerCfgRd  <= '1';            -- And trigger the readout of the EEPROM.
            elsif (GoWrite_i = '1') then
               SerCfgDev <= DeviceAddr_i;   -- Catch the relevent physical device address.
               BusPort   <= ChannelAddr_i;  -- and the bus port that we are interested in
               SerCfgWrt <= '1';            -- and trigger a write operation to the eeprom
            end if;
         else
            SerCfgRd  <= '0';               -- This trigger has done it's job so let it rest.
            SerCfgWrt <= '0';               -- and this little trigger went to market ...
         end if;
      end if;
   end process;

------------------------------------------------------------------------
-- FSM to write data to the EEPROM.
------------------------------------------------------------------------
   BUSFREE_PERIOD <= "00000000000010100000" when (SerCfgDev = EEPROM and WrtRomActive = '1') else "00000000000010100000";   -- period of time between bus transactions = 5.1ms for eeprom
                                                                                                                            --  writes, 1.5us for others.

   SerCfgWriteRom : process (SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1') then     -- First clock after reset true
            WrtRomActive <= '0';
            WrtLoadFlag  <= '0';
            WrtStartReq  <= '0';
            WrtClkEnable <= '0';                -- Kill the clocks
            DataInAck_o <= '0';                 -- Cancel the acknowledge flag
            CfgWrtState <= CfgWrtIdle;
         else
            case CfgWrtState is
               when CfgWrtIdle =>      -- LOOKING FOR A CHANCE TO DO SOMETHING
                  if (SerCfgWrt = '1') then
                     WrtRomActive <= '1';
                     CfgWrtAddr <= (others => '0');     -- Start from address 0.
                     CfgWrtState <= CfgWrtWordLoop;
                  else
                     WrtRomActive <= '0';               -- Go show the world we're idle.
                     WrtClkEnable <= '0';               -- Kill the clocks
                     CfgWrtState <= CfgWrtIdle;
                  end if;

               when CfgWrtWordLoop =>  -- OUTER LOOP FOR WRITE SEQUENCE
                  if (CfgWrtAddr < RdWrtBytes_i) then
                     if (CfgWrtAddr = "0000000") then
                        WrtStartReq <= '1';                 -- Send the start request
                     end if;
                     CfgWrtState <= CfgWrtDataByte;
                  else
                     WrtStopReq <= '1';
                     CfgWrtState <= CfgWrtSendStop;          -- All done, send stop and go home.
                  end if;

               when CfgWrtDataByte =>
                  if (WrtStartReq = '1') then
                     WrtStartReq <= '0';                 -- 1 clock delay for process to bite and cancel start request on first time through.
                     WrtClkEnable <= '1';                -- Start our clocks
                     CfgWrtState <= CfgWrtDataByte;
                  elsif (StartBusy = '0') then              -- Go look for acknowledge
                     if (SdaAck = '1') then                 -- We have received a good acknowledge.
                        WrtLoadData <= DataIn_i;            -- Set up the current data byte
                        DataInAck_o <= '1';                 -- Acknowledge the byte from the source
                        WrtLoadFlag <= '1';                 -- and load the address data.
                        CfgWrtAddr <= CfgWrtAddr + 1;       -- Increment the byte counter
                        CfgWrtState <= CfgWrtDataWait;
                     else
                        CfgWrtState <= CfgWrtIdle;      -- Bad one ... Error !!    
                     end if;
                  else
                     CfgWrtState <= CfgWrtDataByte;                  
                  end if;
                     
               when CfgWrtDataWait =>
                  if (WrtLoadFlag = '1') then
                     WrtLoadFlag <= '0';
                     DataInAck_o <= '0';                 -- Cancel the acknowledge flag
                     CfgWrtState <= CfgWrtDataWait;
                  elsif (WriteBusy = '0') then
                     if (SdaAck = '1') then              -- We have received a good acknowledge.
                        CfgWrtState <= CfgWrtWordLoop;
                     else
                        CfgWrtState <= CfgWrtIdle;      -- Bad one ... Error !!    
                     end if;                                
                  else
                      CfgWrtState <= CfgWrtDataWait;
                  end if;
                  
               when CfgWrtSendStop =>
                  if (WrtStopReq = '1') then
                     WrtStopReq <= '0';
                     WrtDlyCntr <= (others => '0');             -- Get ready for a delay.
                     CfgWrtState <= CfgWrtSendStop;
                  elsif (StopBusy = '0') then
                     CfgWrtState <= CfgWrtDelay;
                  else
                     CfgWrtState <= CfgWrtSendStop;
                  end if;
                   
               when CfgWrtDelay =>     -- DELAY TILL TIME for another operation.
                  WrtClkEnable <= '0';                    -- Kill our clocks for a while
                  if (WrtDlyCntr < BUSFREE_PERIOD) then
                     WrtDlyCntr <= WrtDlyCntr + 1;
                     CfgWrtState <= CfgWrtDelay;
                  else
                     CfgWrtState <= CfgWrtIdle;      -- Go back to idle.
                  end if;
                  
               when others =>
                   CfgWrtState <= CfgWrtIdle;    
            end case;
         end if;
      end if;
   end process;

------------------------------------------------------------------------
-- FSM to read out the EEPROM.
------------------------------------------------------------------------
   SerCfgReadRom : process (SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1') then         -- First clock after reset true
            RdRomActive  <= '0';
            RdClkEnable  <= '0';         -- Kill the clocks
            RdStartReq   <= '0';         -- Kill all the service requests
            RdStopReq    <= '0';
            RdEnableFlag <= '0';         -- Shut down the read process.
            DataOutRdy_o <= '0';         -- Data is valid on the DataOut signals - NOT
            CfgRdState <= CfgRdIdle;
         else
            case CfgRdState is
               when CfgRdIdle =>      -- LOOKING FOR A CHANCE TO DO SOMETHING
                  if (SerCfgRd = '1') then
                     RdRomActive <= '1';
                     CfgRdAddr  <= "0000001";            -- Preload the byte counter for zero origin.
                     RdClkEnable  <= '1';                -- Start the clocks
                     RdStartReq   <= '1';                -- Send the start request
                     CfgRdState <= CfgRdStart;           -- Go wait for results
                  else
                     DataOutRdy_o <= '0';                -- Data is valid on the DataOut signals - NOT
                     RdRomActive <= '0';                 -- Go show the world we're idle.
                     RdClkEnable <= '0';                 -- Kill the clocks
                     CfgRdState <= CfgRdIdle;
                  end if;
               when CfgRdStart =>   -- WAIT FOR THE RESTART ACKNOWLEDGE.
                  if (RdStartReq = '1') then              -- First clock cycle at this state
                     RdStartReq <= '0';                  -- So wait for the output to rationalize and kill the load request meantime.
                     CfgRdState <= CfgRdStart;
                  elsif (StartBusy = '0') then            -- Else wait for the eight bits of address to go out
                     if (SdaAck = '1') then              -- We have received a good acknowledge.
                        RdEnableFlag <= '1';            -- Begin down the read process.
                        AckFlag <= '1';                 -- Set up for normal acknowledge. 
                        CfgRdState <= CfgRdRdDat1;
                     end if;
                  else
                     CfgRdState <= CfgRdStart;
                  end if;
               when CfgRdRdDat1 =>    -- NOW WE BEGIN TO READ IN DATA. SET UP THE PROCESS FOR 254 BYTES TO BE READ.
                  if (CfgRdAddr < RdWrtBytes_i) then
                     if (DataOutAck_i = '1') then
                        DataOutRdy_o <= '0';         -- Cancel dataOutRdy signal on ack
                     end if;
                     if (RdDataRdy = '1') then               -- We have a word .. Stash it in the appropriate place                           
                        DataOut_o <= RdByteValue_4 & RdByteValue_3 & RdByteValue_2 & RdByteValue_1 & RdByteValue_0;
                        DataOutRdy_o <= '1';         -- Data is valid on the DataOut signals
                        CfgRdAddr <= CfgRdAddr + 1;
                     end if;
                     CfgRdState <= CfgRdRdDat1;
                  else
                     AckFlag <= '0';                         -- No acknowledge on this last byte.
                     CfgRdState <= CfgRdRdDat2;
                  end if;
               when CfgRdRdDat2 =>      -- CLEAN UP THE READ PROCESS IN A GOOD WAY
                  if (DataOutAck_i = '1') then
                     DataOutRdy_o <= '0';         -- Cancel dataOutRdy signal on ack
                  end if;
                  if (RdDataRdy = '1') then                   -- finish reading the last byte to be read with no acknowledge sent
                     DataOut_o <= RdByteValue_4 & RdByteValue_3 & RdByteValue_2 & RdByteValue_1 & RdByteValue_0;
                     DataOutRdy_o <= '1';         -- Data is valid on the DataOut signals
                     RdEnableFlag <= '0';                    -- Shut down the read process.
                     RdStopReq <= '1';                       -- that the serial data be loaded.
                     CfgRdState <= CfgRdSendStop;
                  else
                     CfgRdState <= CfgRdRdDat2;
                  end if;                
               when CfgRdSendStop =>
                  if (DataOutAck_i = '1') then
                     DataOutRdy_o <= '0';         -- Cancel dataOutRdy signal on ack
                  end if;
                  if (RdStopReq = '1') then
                     RdStopReq <= '0';
                     RdDlyCntr <= (others => '0');
                     CfgRdState <= CfgRdSendStop;
                  elsif (StopBusy = '0') then
                     CfgRdState <= CfgRdDelay;
                  else
                     CfgRdState <= CfgRdSendStop;
                  end if;
               when CfgRdDelay =>     -- DELAY TILL TIME for another operation.
                  RdClkEnable <= '0';                 -- Kill the clocks
                  if (RdDlyCntr < BUSFREE_PERIOD) then
                     RdDlyCntr <= RdDlyCntr + 1;
                     CfgRdState <= CfgRdDelay;
                  else
                     CfgRdState <= CfgRdIdle;      -- Go back to idle.
                  end if;
                  
               when others =>
                  CfgRdState <= CfgRdIdle;    
            end case;
         end if;
      end if;
   end process;

------------------------------------------------------------------------
-- Start + Select sequence generator for I2C bus
------------------------------------------------------------------------
    SendStartReq <= WrtStartReq or RdStartReq;    

    CfgSendStartProc : process (SysClk_i)
    begin
        if (rising_edge(SysClk_i)) then
            if (Reset_i = '1') then
                StartBusy <= '0';
                StrtLoadFlag <= '0';
                StrtClkEnable <= '0';                    -- and clock resources
                StrtState <= StrtIdle;
            else
                case StrtState is
                    when StrtIdle =>     -- WAITING FOR BUSINESS
                        if (SendStartReq = '1') then
                            StartBusy <= '1';
                            StrtClkEnable <= '1';              -- Start the clocks
                            StrtState <= StrtSync;
                        else
                            StartBusy <= '0';
                            StrtClkEnable <= '0';               -- Kill the clocks
                            StrtState <= StrtIdle;
                        end if;
                    when StrtSync =>
                        if (RdStrobe = '1') then
                            StrtLoadData <= (others => '0');   -- Send the start bit by requesting 
                            StrtLoadFlag <= '1';               -- that the serial data be loaded.
                            StrtState <= StrtStart;
                        else
                            StrtState <= StrtSync;
                        end if;    
                    when StrtStart =>       -- WAITING FOR START SEQUENCE TO FINISH
                        StrtLoadFlag <= '0';                    -- Kill the load request.
                        if (ClkPhase0 = '1') then               -- We've got the start sequence correct
                            StrtLoadData <= SerCfgDev & not WriteFlag; -- Load the select code in first.
                            StrtLoadFlag <= '1';
                            StrtState <= StrtSendSlct;
                        else
                            StrtState <= StrtStart;
                        end if;
                    when StrtSendSlct =>    -- WAIT UNTIL ALL DEVICE SELECT BITS HAVE BEEN WRITTEN and ACKNOWLEDGE READ.
                        if (StrtLoadFlag = '1') then
                            StrtLoadFlag <= '0';
                        elsif (WriteBusy = '0') then
                            StrtState <= StrtIdle;
                        else
                            StrtState <= StrtSendSlct;
                        end if;    
                    when others =>
                        StrtState <= StrtIdle;
                end case;
            end if;
        end if;
    end process;

    CfgSendStartWrtFlagProc : process (SysClk_i)
    begin
        if (rising_edge(SysClk_i)) then
            if (WrtStartReq = '1') then
                WriteFlag <= '1';   --- WrtSndStartWriteFlag;
            elsif (RdStartReq = '1') then
                WriteFlag <= '0';
            end if;
        end if;
    end process;

------------------------------------------------------------------------
-- Stop sequence generator for I2C bus
------------------------------------------------------------------------
    SendStopReq <= WrtStopReq or RdStopReq;

    CfgSendStopProc : process (SysClk_i)
    begin
        if (rising_edge(SysClk_i)) then
            if (Reset_i = '1') then
                StopBusy <= '0';
                StpLoadFlag <= '0';
                StpClkEnable <= '0';                    -- and clock resources
                SndStopState <= SndStopIdle;
            else
                case SndStopState is
                    when SndStopIdle =>
                        StpLoadFlag <= '0';
                        if (SendStopReq = '1') then
                            StopBusy <= '1';
                            StpClkEnable <= '1';
                            SndStopState <= SndStopStart;
                        else
                            StopBusy <= '0';
                            StpClkEnable <= '0';
                            SndStopState <= SndStopIdle;
                        end if;
                    when SndStopStart =>
                        if (ClkPhase0 = '1') then
                            StpLoadData <= (others => '0');        -- Send the stop bit by requesting
                            StpLoadFlag <= '1';
                            SndStopState <= SndStopStop;
                        else
                            SndStopState <= SndStopStart;
                        end if;
                    when SndStopStop =>
                        StpLoadFlag <= '0';                         -- Cancel the load flag.
                        if (RdStrobe = '1') then                    -- Wait for the appropriate moment
                            StpLoadData <= (others => '1');         -- Send the stop bit now after the clock has been high for > 600ns. 
                            StpLoadFlag <= '1';                     -- Load it.
                            SndStopState <= SndStopIdle;            -- Go do it all again ... someday !    
                        else
                            SndStopState <= SndStopStop;
                        end if;
                    when others =>
                        SndStopState <= SndStopIdle;
                end case;
            end if;
        end if;
    end process;

------------------------------------------------------------------------
-- Clock generator for read / write FSMs
------------------------------------------------------------------------
    ClkPhase0 <=  not (SerCfgClk(0) or SerCfgClk(1));    -- Clock state low 
    ClkPhase1 <= (not SerCfgClk(1)) and SerCfgClk(0);    -- Clock state high
    ClkPhase2 <= (not SerCfgClk(0)) and SerCfgClk(1);    -- Clock state low
    ClkEnableFlag <= StpClkEnable or StrtClkEnable or RdClkEnable or WrtClkEnable;

    CfgClkGenProc : process (SysClk_i)
    begin
        if (rising_edge(SysClk_i)) then
            if (ClkEnableFlag = '1') then
                if (ClkCntr = CLK_PERIOD) then
                    ClkCntr <= (others => '0');         -- Time we changed the clock state
                    if (SerCfgClk(1) = '1') then        -- test the clock phase for end of cycle
                        SerCfgClk <= (others => '0');
                    else
                        SerCfgClk <= SerCfgClk + 1;     -- Generate the clock triplet
                    end if;
                else
                    if (ClkPhase1 = '1' and ClkCntr = RD_STROBE) then
                        RdStrobe <= '1';
                    else
                        RdStrobe <= '0';
                    end if;
                    ClkCntr <= ClkCntr + 1;
                end if;
            else
                ClkCntr <= (others => '0');             -- We're initially sync'ed to the start of any FSM through RdRomActive.
                SerCfgClk <= "01";                      -- Initial clock state is high.
                RdStrobe <= '0';                        -- Cancell any hung read strobe ... Murphy lurks here.
            end if;
        end if;
    end process;

    MstrScl_o <= SerCfgClk(0);                                -- Output the clock state.

------------------------------------------------------------------------
-- Timing and management for parallel to serial converter to SDA
------------------------------------------------------------------------
    LoadData <= WrtLoadData  when (WrtLoadFlag = '1')  else 
                StpLoadData  when (StpLoadFlag = '1')  else
                StrtLoadData when (StrtLoadFlag = '1') else 
                x"FF";

    LoadFlag  <= WrtLoadFlag or StpLoadFlag or StrtLoadFlag;
    SdaEnable <= WrtSdaEnable or AckSdaEnable;
    MstrWrtEn_o  <= SdaEnable;

    MstrSda_o  <= not AckFlag when (AckSdaEnable = '1') else SerWrtData(7);  -- Output the serial data to the world.

    CfgWrtDataProc : process (SysClk_i)
    begin
        if (rising_edge(SysClk_i)) then
            if (LoadFlag = '1') then
                SerWrtData <= LoadData;
                BitCntr <= (others => '0');
                DataArm <= '0';
            else
                case SerCfgClk is
                    when "00" =>
                        if (DataArm = '1') then
                            SerWrtData <= SerWrtData(6 downto 0) & '1';
                            BitCntr <= BitCntr + 1;
                            DataArm <= '0';
                        end if;
                    when "01" => 
                        null;
                    when "10" =>
                        DataArm <= '1';
                    when others =>
                        null;
                end case;
            end if;
        end if;
    end process;

   SdaAckBit <= (not MstrSda_i(0) or not MstrSda_i(1) or not MstrSda_i(2) or not MstrSda_i(3) or not MstrSda_i(4)); -- even if one device responds, give a good acknowledge

   
    CfgWrtDataStateProc : process (SysClk_i)
    begin
        if (rising_edge(SysClk_i)) then
            if (Reset_i = '1') then
                WriteBusy <= '0';
                WrtSdaEnable <= '0';                    -- Release hold on sda line
                WrtDataState <= WrtDataIdle;
            else
                case WrtDataState is
                    when WrtDataIdle =>
                        if (LoadFlag = '1') then
                            WriteBusy <= '1';
                            IfcWrtErr_o <= (others => '0');         -- Reset any previous acknowledge errors.
                            IfcRdErr_o  <= (others => '0');
                            WrtDataState <= WrtDataBusy;
                            WrtSdaEnable <= '1';        -- Enable the sda line.
                        else
                            WriteBusy <= '0';
                            WrtSdaEnable <= '0';        -- Disable the sda line when idle.
                            WrtDataState <= WrtDataIdle;
                        end if;
                    when WrtDataBusy =>
                     if (BitCntr(2 downto 0) = "111" and ClkPhase2 = '1') then      -- Ninth bit time ... We're due an acknowledgement bit
                            WrtSdaEnable <= '0';        -- Disable the master SDA sig and listen for an acknowledge
                            WrtDataState <= WrtDataAcknowledge;
                        else
                            WrtDataState <= WrtDataBusy;
                        end if;
                    when WrtDataAcknowledge =>  -- WAIT FOR ACKNOWLEDGE TIME AND SAMPLE SDA
                        if (RdStrobe = '1') then        -- Read the acknowledge result in the middle of the clock high pulse
                            SdaAck <= SdaAckBit;
                            DataValidPort <= not MstrSda_i and BusPort;
                            if (WrtRomActive = '1') then
                            IfcWrtErr_o <= MstrSda_i and BusPort; -- lock in any errors on the acknowledge
                            else
                            IfcRdErr_o <= MstrSda_i and BusPort; -- lock in any errors on the acknowledge
                            end if;
                            WrtDataState <= WrtDataEndFrame;
                        else
                            WrtDataState <= WrtDataAcknowledge;
                        end if;
                    when WrtDataEndFrame =>     -- WAIT FOR END OF NINTH BIT TIME.
                        if (ClkPhase0 = '1') then
                            WrtDataState <= WrtDataIdle;
                        else
                            WrtDataState <= WrtDataEndFrame;
                        end if;    
                    when others =>
                        WrtDataState <= WrtDataIdle;
                end case;
            end if;
        end if;
    end process;
            
------------------------------------------------------------------------
-- Timing and management for serial to parallel converter from SDA
------------------------------------------------------------------------    
    SerCfgRdData : process (SysClk_i)
    begin
        if (rising_edge(SysClk_i)) then
            if (RdEnableFlag = '1') then
                if (RdBitCntr < BITS_PER_BYTE) then     -- HAVE WE COLLECTED ALL THE BITS YET ?
                    AckSdaEnable <= '0';                             -- Release the serial signal to the device
                    RdDataRdy <= '0';                               -- Cancel any previous data ready flag.
                    RdAckActive <= '0';                             -- and local flag used during acknowledgement.
                    if (RdStrobe = '1') then                 -- IT'S TIME TO READ THE SDA DATA BIT
                        SerRdData_0(0) <= MstrSda_i(0) and DataValidPort(0);             -- Catch the SDA level at the correct time
                        SerRdData_1(0) <= MstrSda_i(1) and DataValidPort(1);             -- Catch the SDA level at the correct time
                        SerRdData_2(0) <= MstrSda_i(2) and DataValidPort(2);             -- Catch the SDA level at the correct time
                        SerRdData_3(0) <= MstrSda_i(3) and DataValidPort(3);             -- Catch the SDA level at the correct time
                        SerRdData_4(0) <= MstrSda_i(4) and DataValidPort(4);             -- Catch the SDA level at the correct time                       
                        ShiftFlag <= '1';                           -- and prepare for a shift opperation.
                        RdBitCntr <= RdBitCntr + 1;                 -- Increment the bit counter.
                    elsif (ShiftFlag = '1') then            -- WE'VE COLLECTED, NOW SHIFT IT.
                        SerRdData_0 <= SerRdData_0(6 downto 0) & '0';   -- After grabbing the bit, shift it through
                        SerRdData_1 <= SerRdData_1(6 downto 0) & '0';   -- After grabbing the bit, shift it through
                        SerRdData_2 <= SerRdData_2(6 downto 0) & '0';   -- After grabbing the bit, shift it through
                        SerRdData_3 <= SerRdData_3(6 downto 0) & '0';   -- After grabbing the bit, shift it through
                        SerRdData_4 <= SerRdData_4(6 downto 0) & '0';   -- After grabbing the bit, shift it through
                        ShiftFlag <= '0';                           -- and make sure it's only one shift.
                    end if;
                else
                    if (ClkPhase0 = '1' and AckSdaEnable = '0') then -- Okay lets start the Acknowledge to the device.
                        AckSdaEnable <= '1';                         -- Enable the tristate gate on the SDA line which sets acknowledge true
                    elsif (ClkPhase2 = '1' and AckSdaEnable = '1') then
                        RdAckActive <= '1';                         -- Wait for the 2nd clock phase to pass.
                    elsif (ClkPhase0 = '1' and RdAckActive = '1') then
                        AckSdaEnable <= '0';
                        RdByteValue_0 <= SerRdData_0;                  -- Latch the data to our correspondents (redundent really, this thing is so slow !)
                        RdByteValue_1 <= SerRdData_1;                  -- Latch the data to our correspondents (redundent really, this thing is so slow !)
                        RdByteValue_2 <= SerRdData_2;                  -- Latch the data to our correspondents (redundent really, this thing is so slow !)
                        RdByteValue_3 <= SerRdData_3;                  -- Latch the data to our correspondents (redundent really, this thing is so slow !)
                        RdByteValue_4 <= SerRdData_4;                  -- Latch the data to our correspondents (redundent really, this thing is so slow !)
                        SerRdData_0 <= (others => '0');               -- Should be redundent.
                        SerRdData_1 <= (others => '0');               -- Should be redundent.
                        SerRdData_2 <= (others => '0');               -- Should be redundent.
                        SerRdData_3 <= (others => '0');               -- Should be redundent.
                        SerRdData_4 <= (others => '0');               -- Should be redundent.
                        RdBitCntr <= (others => '0');               -- Go back and get a another byte
                        RdDataRdy <= '1';                           -- Flag our sponsor that the data is ready.
                    end if;
                end if;
            else                                        -- NOT READING ... BUT BE READY.
                RdDataRdy <= '0';                                -- Hold off the data ready flag.
                ShiftFlag <= '0';
                SerRdData_0 <= (others => '0');                    -- Start clean ... (Hey mister clean ... Your dirty now too ! NY 2002)
                SerRdData_1 <= (others => '0');                    -- Start clean ... (Hey mister clean ... Your dirty now too ! NY 2002)
                SerRdData_2 <= (others => '0');                    -- Start clean ... (Hey mister clean ... Your dirty now too ! NY 2002)
                SerRdData_3 <= (others => '0');                    -- Start clean ... (Hey mister clean ... Your dirty now too ! NY 2002)
                SerRdData_4 <= (others => '0');                    -- Start clean ... (Hey mister clean ... Your dirty now too ! NY 2002)
                RdBitCntr <= (others => '0');
                AckSdaEnable <= '0';                             -- Allow others to use the SDA signal.
            end if;
        end if;
    end process;

end Behavioral;