------------------------------------------------------------------------
-- Copyright Notice:
-- (c) Copyright 2004 - 2008, National Optical Astronomy Observatory,
-- which is operated for the National Science Foundation by the
-- Association of Universities for Research in Astronomy under
-- Cooperative Agreement No. AST-0132798.
------------------------------------------------------------------------
-- PROJECT      :   MONSOON-TORRENT
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   CFG_SERVICES
-- MODULE       :   I2C MANAGER
-- VERSION      :   2.02
------------------------------------------------------------------------
-- FILE NAME   :   I2C_Manager_V203.vhd
-- AUTHOR        :   Peter Moore
-- DATE            :   1/23/2011 -- filled out the Vbb shell
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.03
------------------------------------------------------------------------
-- DESCRIPTION  :  Built on top of the I2C interface, this module controls the higher level read write functions
--  to extract telemetry and eeprom data.                  
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
-- eepDataRdWrtAddr_o designations
-- Value                 Register                 Value            Register
-- "00000"     eepDataRegister00    "10000"             None
-- "00001"     eepDataRegister01    "10001"            None
-- "00010"     eepDataRegister02    "10010"            None
-- "00011"     eepDataRegister03    "10011"            None 
-- "00100"     eepDataRegister04    "10100"            None
-- "00101"     eepDataRegister05    "10101"            None 
-- "00110"     eepDataRegister06    "10110"             None
-- "00111"     eepDataRegister07    "10111"            None 
-- "01000"     eepDataRegister08    "11000"            None
-- "01001"     eepDataRegister09    "11001"            None 
-- "01010"     eepDataRegister10    "11010"            None 
-- "01011"     eepDataRegister11    "11011"            None 
-- "01100"     eepDataRegister12    "11100"            None 
-- "01101"     eepDataRegister13    "11101"             None
-- "01110"     eepDataRegister14    "11110"            None 
-- "01111"     eepDataRegister15    "11111"            None 
-- Temperature selection for DHE tempertaure servo is:
--  DheTempSensorSlct_reg 
--                  Value                                   Sensor Selected for servo reference
--                  "000"                                   fpgaTemp (From system monitor)
--                  "001"                                   LCB_temperature0  (From CFG module)
--                  "010"                                   LCB_temperature1  (From CFG module)
--                  "011"                                   AFE0_temperature  (From CFG module)
--                  "100"                                  AFE0_temperature1  (From CFG module)
--                  "101"                                   AFE1_temperature0  (From CFG module)
--                  "110"                                   AFE1_temperature1  (From CFG module)
--                  "111"                                    Manual seting from PAN (for testing purposes)

------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	     AUTHOR  VERSION  MODIFICATIONS
-- 06/29/2009  PCM     1.00    INITIAL VERSION
-- 02/04/2010  PCM     1.02    REMOVED WISHBONE CYCLE REQUEST LOGIC - WE WILL JUST PIPE TEN BITS OF DATA DIRECTLY
--                                                  TO THE PSM FOR TEMP SERVO CONTROL RATHER THAN GO THROUGH WISHBONE BUS CYCLE
-- 02/18/2010   PCM     1.03    CHANGED EEPROM ADDRESS TO REFECT NEW HARDWARE SCHEME TO INCORPORATE THE SILICON SERIAL NUMBER
-- 03/22/2010  PCM     1.04     Reduced Command word width from 31 bits to 17 bits to reduce warning messages
-- 12/14/2010   PCM     2.02     Added Vbb voltage telemetry output for PSM Rev -A- and TSM-Utility Rev -A- board 
-- 10/18/2011   PCM      2.10     Changed some signals to constants (adcReadData, tempReadData, ssnReadData, eepWriteBytes). 
-----------------------------------------------------------------------
-- PORT DEFINITIONS :
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity I2C_Manager_Ver210 is
	Port (
      SysClk_i     : in  std_logic;
      msClk_i      : in  std_logic;    -- One SysClk pulse every millisecond.
      Reset_i      : in  std_logic;
-- Module Control
      ChannelAddr_o: out  std_logic_vector(4 downto 0);  -- Bit select address of module to send / receive data -- note that more than one port can be selected
      DeviceAddr_o : out std_logic_vector(6 downto 0);   -- Peripheral bus address
      GoRead_o     : out std_logic; -- Initiate a read operation from the addressed bus address.
      GoWrite_o    : out std_logic; -- Initiate a write operation to the addressed bus address
      RdWrtBytes_o : out std_logic_vector(6 downto 0);  -- Number of bytes to read or write to the bus device  (i.e. max 127 bytes per transaction).
-- Read Data interface
      DataOut_i    : in std_logic_vector(39 downto 0); -- Data read from the bus device (8 bits x 5 ports)
      DataOutRdy_i : in  std_logic;    -- Data is valid on the DataOut signals
      DataOutAck_o : out std_logic;    -- Acknowledge from controller upon accepting the byte
-- Write data interface
      DataIn_o     : out std_logic_vector(7 downto 0); -- Data to write to the bus device      
      DataInRdy_o  : out std_logic;    -- Write data is valid on the DataIn signals
      DataInAck_i  : in  std_logic;    -- Acknowledge from this interface upon accepting the byte
-- I2C interface status signals
      IfcBusy_i    : in  std_logic;
      IfcRdErr_i   : in  std_logic_vector(4 downto 0); -- An error occured while reading the bus device (no good acknowledge)
      IfcWrtErr_i  : in  std_logic_vector(4 downto 0); -- An error occured while writing to the bus device (no good acknowledge)
-- I2C Mux control signals 
      DetectMods_o : out std_logic;   -- pulse to command the mux to go through a bus detect function to detect connected I2C busses that are connected.
-- I2C Mux Status inputs
      PSM_Detect_i  : in  std_logic;   -- PSM I2C bus detected
      TSM_Detect_i  : in  std_logic;
      AFE0_Detect_i : in  std_logic;
      AFE1_Detect_i : in  std_logic;
-- Cfg Command strobes
      DetectI2C_i          : in   std_logic;  -- Pulsed to initiate detection of I2C buses
-- Temperature scanner Control signals
      tempScanEnable_i     : in   std_logic;   -- Enable temperature scanning and broadcasts
      tempScanRate_i       : in   std_logic_vector(9 downto 0);   -- Period between temperature sensor scans when scanning enabled - in units milliseconds
-- EEPROM register data control
      eepCmdReg_i          : in   std_logic_vector(16 downto 0);  -- command register - bit-16=readEEprom, bit-15=WriteEEprom, bit-14=Read Temps, bit-13=read SSN, bits(12:8)=Channel Address, bits(7:0)=Page address.
      eepCmdStatus_o       : out  std_logic_vector(3 downto 0);   -- Status for eeprom reads/writes - Bits(3:2)=read, bits(1:0)= write, "00" = not busy + no error, "01" = not busy + error, "10" = busy, "11" = busy + error.
      eepDataRdStb_o       : out  std_logic;  -- Pulsed from I2C manager to read write data writen by PAN to the register control into the I2C manager module
      eepDataWrtStb_o      : out  std_logic;  -- Pulsed from I2C manager to write data read from eeproms to the register control module
      eepDataRdWrtAddr_o   : out  std_logic_vector(4 downto 0); -- Address from the I2C manager to select which eepDataRegister to select for reading or writing.
      eepDataRegister_io   : inout std_logic_vector(31 downto 0);  -- Sixteen data registers for reading / writing data from/to EEProms
-- Module data from I2C temperature and SSN bus reads
      LCB_temperature0_o   : out  std_logic_vector(11 downto 0);  -- Temperature of module from first sensor (address 0x48)
      LCB_temperature1_o   : out  std_logic_vector(11 downto 0);  -- Temperature of module from second sensor (address 0x49)
      PSM_temperature0_o   : out  std_logic_vector(11 downto 0);
      PSM_temperature1_o   : out  std_logic_vector(11 downto 0);
      TSM_temperature0_o   : out  std_logic_vector(11 downto 0);
      TSM_temperature1_o   : out  std_logic_vector(11 downto 0);
      AFE0_temperature0_o  : out  std_logic_vector(11 downto 0);
      AFE0_temperature1_o  : out  std_logic_vector(11 downto 0);
      AFE1_temperature0_o  : out  std_logic_vector(11 downto 0);
      AFE1_temperature1_o  : out  std_logic_vector(11 downto 0);
      LCB_SSN_o            : out  std_logic_vector(31 downto 0);  -- Silicon serial number of LCB module at address 0x50
      PSM_SSN_o            : out  std_logic_vector(31 downto 0);
      TSM_SSN_o            : out  std_logic_vector(31 downto 0);
      AFE0_SSN_o           : out  std_logic_vector(31 downto 0);
      AFE1_SSN_o           : out  std_logic_vector(31 downto 0);
      VbbVoltageTel_o      : out  std_logic_vector(11 downto 0) 
      );
end I2C_Manager_Ver210;

architecture Behavioral of I2C_Manager_Ver210 is
   constant TEMP0_ADDR  : std_logic_vector(6 downto 0) := "1001000";  -- Temp sensor 0 I2C address = 0x48
   constant TEMP1_ADDR  : std_logic_vector(6 downto 0) := "1001001";  -- Temp sensor 1 I2C Address = 0x49
--   constant SSN_ADDR    : std_logic_vector(6 downto 0) := "1010000";  -- Silicon Serial number I2C Address = 0x56 -- Prototype baords - Disabled until we get SSNs !!
--   constant EEPROM_ADDR : std_logic_vector(6 downto 0) := "1010000";  -- EEProm I2C address = 0x50 -- For prototype hardware which is missing AD2 bit tied high.
   constant SSN_ADDR    : std_logic_vector(6 downto 0) := "1010000";  --   -- Silicon Serial number I2C Address = 0x50 -- Production hardware
   constant EEPROM_ADDR : std_logic_vector(6 downto 0) := "1010100";  -- EEProm I2C address = 0x54 -- For production hardware
   constant ADC_ADDR    : std_logic_vector(6 downto 0) := "0001000";  -- VBB voltage telemetry ADC I2C address = 0x08 -- for production hardware

   constant LCB_CHAN    : std_logic_vector(4 downto 0) := "00001";    -- Local control Board channel address
   constant PSM_CHAN    : std_logic_vector(4 downto 0) := "00010";    -- Power Supply Module channel address
   constant TSM_CHAN    : std_logic_vector(4 downto 0) := "00100";    -- Transition Module channel address
   constant AFE1_CHAN   : std_logic_vector(4 downto 0) := "01000";    -- AFE 1 channel address
   constant AFE2_CHAN   : std_logic_vector(4 downto 0) := "10000";    -- AFE 2 channel address
   
   constant ADC_SCAN_RATE : std_logic_vector(9 downto 0) := "0000111111"; -- 63 millisecond scan of the Vbb voltage 
                                                                      
   type eepReadStateType is (eepReadIdle, eepRead_1, eepRead_2, eepRead_3, eepRead_4, eepRead_5, eepRead_6, eepRead_7);
   signal eepReadState : eepReadStateType;  
   type eepWriteStateType is (eepWriteIdle, eepWrite_1, eepWrite_2, eepWrite_3, eepWrite_4, eepWrite_5, eepWrite_6);
   signal eepWriteState : eepWriteStateType;
   type tempReadStateType is (tempReadIdle, tempRead_1, tempRead_2, tempRead_3, tempRead_4);
   signal tempReadState : tempReadStateType;  
   type ssnReadStateType is (ssnReadIdle, ssnRead_1, ssnRead_2, ssnRead_3, ssnRead_4, ssnRead_5, ssnRead_6);
   signal ssnReadState : ssnReadStateType;
   type adcReadStateType is (adcReadIdle, adcRead_1, adcRead_2, adcRead_3, adcRead_4);
   signal adcReadState : adcReadStateType;  
  
   signal eepMemAddr : std_logic_vector (15 downto 0);
   signal I2C_DeviceAddr : std_logic_vector(6 downto 0);
   signal I2C_Channel, I2C_Channelmask : std_logic_vector(4 downto 0);
   signal eepReadActive, eepWriteActive, tempReadActive, ssnReadActive, adcReadActive : std_logic;
   signal eepReadGo, eepWriteGo, tempReadGo, ssnReadGo, adcReadGo : std_logic;
   signal ModRstFlag, I2C_busy, eepCmdPendingFlag, eepCmdRegRdy : std_logic;
   signal eepCmdPending, eepCommand : std_logic_vector(16 downto 0);
   signal eepReadError, eepWriteError : std_logic;
   signal I2C_ChanRdCheck : std_logic;
   signal DataOutByte : std_logic_vector(7 downto 0);
   
-- Temperature sensor periodic readout sigs
   signal tempReadReq : std_logic;
   signal tempPerCntr : std_logic_vector(9 downto 0);
   signal tempChanAddrToggle : std_logic;
-- Adc sensor periodic readout sigs
   signal adcReadReq : std_logic;
   signal adcPerCntr : std_logic_vector(9 downto 0);
-- I2C mux signals   
   signal eepReadWrtStrobe, eepWriteWrtStrobe, tempReadWrtStrobe, ssnReadWrtStrobe, adcReadWrtStrobe : std_logic;
   signal eepReadRdStrobe, tempReadRdStrobe, ssnReadRdStrobe, adcReadRdStrobe : std_logic;
   signal eepReadBytes, tempReadBytes, ssnReadBytes, adcReadBytes : std_logic_vector(6 downto 0);
   signal eepReadData, eepWriteData : std_logic_vector(7 downto 0);
   signal eepReadDataRdy, eepWriteDataRdy, tempReadDataRdy, ssnReadDataRdy, adcReadDataRdy : std_logic;
   signal eepReadDataOutAck, tempReadDataOutAck, ssnReadDataOutAck, adcReadDataOutAck : std_logic;
   constant adcReadData : std_logic_vector(7 downto 0) := "00001000"; -- write to the config register to start aconversion. Config asks for unipolar mode
   constant ssnReadData : std_logic_vector(7 downto 0) := "00000001"; -- ssn registers start at 1
   constant tempReadData: std_logic_vector(7 downto 0) := "00000000"; -- temperature data register is zero
   constant eepWriteBytes : std_logic_vector(6 downto 0) := "1000010"; -- equals 64 bytes + 2 address bytes
-- Register control signals
   signal eepReadRegAddr, eepWriteRegAddr : std_logic_vector(4 downto 0);
   signal eepReadRegWrtStrobe, eepWriteRegRdStrobe : std_logic;
   signal eepWriteWord : std_logic_vector(31 downto 0);
   signal eepReadWord  : std_logic_vector(23 downto 0);
-- Temperature registers
   signal LCB_temperature0   : std_logic_vector(11 downto 0);  -- Temperature of module from first sensor (address 0x48)
   signal LCB_temperature1   : std_logic_vector(11 downto 0);  -- Temperature of module from second sensor (address 0x49)
   signal PSM_temperature0   : std_logic_vector(11 downto 0);
   signal PSM_temperature1   : std_logic_vector(11 downto 0);
   signal TSM_temperature0   : std_logic_vector(11 downto 0);
   signal TSM_temperature1   : std_logic_vector(11 downto 0);
   signal AFE0_temperature0  : std_logic_vector(11 downto 0);
   signal AFE0_temperature1  : std_logic_vector(11 downto 0);
   signal AFE1_temperature0  : std_logic_vector(11 downto 0);
   signal AFE1_temperature1  : std_logic_vector(11 downto 0);
   signal LCB_SSN            : std_logic_vector(31 downto 0);  -- Silicon serial number of LCB module at address 0x50
   signal PSM_SSN            : std_logic_vector(31 downto 0);
   signal TSM_SSN            : std_logic_vector(31 downto 0);
   signal AFE0_SSN           : std_logic_vector(31 downto 0);
   signal AFE1_SSN           : std_logic_vector(31 downto 0);
   signal ADC_voltage        : std_logic_vector(11 downto 0);

begin
   
   DetectMods_o         <= DetectI2C_i;
   LCB_temperature0_o   <= LCB_temperature0;
   LCB_temperature1_o   <= LCB_temperature1;
   PSM_temperature0_o   <= PSM_temperature0;
   PSM_temperature1_o   <= PSM_temperature1;
   TSM_temperature0_o   <= TSM_temperature0;
   TSM_temperature1_o   <= TSM_temperature1;
   AFE0_temperature0_o  <= AFE0_temperature0;
   AFE0_temperature1_o  <= AFE0_temperature1;
   AFE1_temperature0_o  <= AFE1_temperature0;
   AFE1_temperature1_o  <= AFE1_temperature1;
   LCB_SSN_o            <= LCB_SSN;
   PSM_SSN_o            <= PSM_SSN;
   TSM_SSN_o            <= TSM_SSN;
   AFE0_SSN_o           <= AFE0_SSN;
   AFE1_SSN_o           <= AFE1_SSN;
   VbbVoltageTel_o      <= ADC_voltage;
   
-- Command processor
   eepCommand <= eepCmdReg_i when (eepCmdPendingFlag = '0') else eepCmdPending;  -- Mux either the pan command or a captured command caught while we were busy.
   eepCmdRegRdy    <= eepCmdReg_i(16) or eepCmdReg_i(15) or eepCmdReg_i(14) or eepCmdReg_i(13);
   I2C_Channelmask <= AFE1_Detect_i & AFE0_Detect_i & TSM_Detect_i & PSM_Detect_i & '1';
   I2C_Busy <= eepReadGo or eepWriteGo or tempReadGo or ssnReadGo or eepReadActive or eepWriteActive or tempReadActive or ssnReadActive or IfcBusy_i;
   eepReadError  <= (IfcRdErr_i(4) and I2C_Channelmask(4)) or (IfcRdErr_i(3) and I2C_Channelmask(3)) or
                    (IfcRdErr_i(2) and I2C_Channelmask(2)) or (IfcRdErr_i(1) and I2C_Channelmask(1)) or (IfcRdErr_i(0) and I2C_Channelmask(0));
   eepWriteError <= (IfcWrtErr_i(4) and I2C_Channelmask(4)) or (IfcWrtErr_i(3) and I2C_Channelmask(3)) or
                    (IfcWrtErr_i(2) and I2C_Channelmask(2)) or (IfcWrtErr_i(1) and I2C_Channelmask(1)) or (IfcWrtErr_i(0) and I2C_Channelmask(0));
   eepCmdStatus_o <= (eepReadGo or eepReadActive or eepCmdPending(16)) & eepReadError & (eepWriteGo or eepWriteActive or eepCmdPending(15)) & eepWriteError;
   
   I2CCommandProc : process(SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1') then
            eepCmdPendingFlag <= '0';
            eepReadGo  <= '0';
            eepWriteGo <= '0';
            tempReadGo <= '0';
            ssnReadGo  <= '0';
            ModRstFlag <= '1';
         elsif (I2C_Busy = '0') then
            case eepCommand(16 downto 13) is
               when "1000" =>   -- Pan command to read eeprom page           
                  eepReadGo <= '1';
                  I2C_DeviceAddr <= EEPROM_ADDR;
                  I2C_Channel <= eepCommand(12 downto 8) and I2C_Channelmask;  -- Prevent non-existant channels from being used
                  eepMemAddr  <= "00" & eepCommand(7 downto 0) & "000000";  -- Convert pages to bytes by x 64, max address of 16383 (0x3FFF) for 24AA128
               when "0100" =>   -- Pan command to write eeprom page
                  eepWriteGo <= '1';
                  I2C_DeviceAddr <= EEPROM_ADDR;
                  I2C_Channel <= eepCommand(12 downto 8) and I2C_Channelmask;  -- Prevent non-existant channels from being used
                  eepMemAddr  <= "00" & eepCommand(7 downto 0) & "000000";  -- Convert pages to bytes by x 64, max address of 16383 (0x3FFF) for 24AA128
               when "0010" =>   -- Pan command to read temperatures 
                  tempReadGo <= '1';
                  I2C_Channel <= eepCommand(12 downto 8); -- and I2C_Channelmask;  -- Prevent non-existant channels from being used
                  if (eepCommand(0) = '0') then
                     I2C_DeviceAddr <= TEMP0_ADDR;
                  else
                     I2C_DeviceAddr <= TEMP1_ADDR;
                  end if;
               when "0001" =>   -- Pan command to read silicon serial numbers 
                  ssnReadGo <= '1';            
                  I2C_DeviceAddr <= SSN_ADDR;
                  I2C_Channel <= eepCommand(12 downto 8) and I2C_Channelmask;  -- Prevent non-existant channels from being used
               when others =>
                  if (ModRstFlag = '1') then  -- First pass through here after reset and module detect finished i.e. ifcBusy = '0';
                     ssnReadGo <= '1';            
                     I2C_DeviceAddr <= SSN_ADDR;
                     I2C_Channel <= I2C_Channelmask;  --  Scan all existant channels for a serial number after reset
                  elsif (tempReadReq = '1') then
                     tempReadGo <= '1';
                     I2C_Channel <= I2C_Channelmask;  -- Scan all existant channels
                     if (tempChanAddrToggle = '0') then
                        I2C_DeviceAddr <= TEMP0_ADDR;
                     else
                        I2C_DeviceAddr <= TEMP1_ADDR;
                     end if;
                  elsif (adcReadReq = '1') then
                     adcReadGo <= '1';
                     I2C_Channel <= TSM_CHAN and I2C_Channelmask;  -- Scan the Vbb ADC on the tsm hardware module
                     I2C_DeviceAddr <= ADC_ADDR;
                  end if;
            end case;
            eepCmdPendingFlag <= '0';
            eepCmdPending <= (others => '0');
         else
            if (eepCmdRegRdy = '1' and eepCmdPendingFlag = '0') then   -- Capture a pan command pending the end of current operation
               eepCmdPending <= eepCmdReg_i;
               eepCmdPendingFlag <= '1';
            end if;
            if (eepReadGo = '1' and eepReadActive = '1') then -- Cancel the read trigger once the fsm is active
               eepReadGo <= '0';
            end if;
            if (eepWriteGo = '1' and eepWriteActive = '1') then -- Cancel the write trigger once the fsm is active
               eepWriteGo <= '0';
            end if;
            if (tempReadGo = '1' and tempReadActive = '1') then -- Cancel the temperature read trigger once the fsm is active
               tempReadGo <= '0';
            end if;
            if (ssnReadGo = '1' and ssnReadActive = '1') then -- Cancel the ssn read trigger once the fsm is active
               ModRstFlag <= '0';  -- Cancel the on reset serial number process trigger
               ssnReadGo <= '0';
            end if;
            if (adcReadGo = '1' and adcReadActive = '1') then -- Cancel the Vbb adc read trigger once the fsm is active
               adcReadGo <= '0';
            end if;
         end if;
      end if;
   end process;

-- Temp sensor periodic readout trigger
   TempPeriodProc : process(SysClk_i)  -- Note that a complete scan of all temp sensors takes two cycles !!
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1' or tempScanEnable_i = '0') then
            tempReadReq <= '0';  -- hold off any requests to the command processor
            TempChanAddrToggle <= '0';
            tempPerCntr <= "0000000010";
         elsif (tempPerCntr >= tempScanRate_i) then
            tempPerCntr <= "0000000010";
            tempReadReq <= '1';
         elsif (msClk_i = '1') then
            tempPerCntr <= tempPerCntr + 1;
         elsif (tempReadReq = '1' and tempReadActive = '1') then
            tempChanAddrToggle <= not tempChanAddrToggle;
            tempReadReq <= '0';
         end if;
      end if;
   end process;

-- Vbb ADC periodic readout trigger
   adcPeriodProc : process(SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1') then
            adcReadReq <= '0';  -- hold off any requests to the command processor
            adcPerCntr <= (others => '0');
         elsif (adcPerCntr >= ADC_SCAN_RATE) then
            adcPerCntr <= (others => '0');
            adcReadReq <= '1';
         elsif (msClk_i = '1') then
            adcPerCntr <= adcPerCntr + 1;
         elsif (adcReadReq = '1' and adcReadActive = '1') then
            adcReadReq <= '0';
         end if;
      end if;
   end process;

   
-- I2C signal Mux
   ChannelAddr_o <= I2C_Channel;
   DeviceAddr_o  <= I2C_DeviceAddr;
   GoWrite_o     <= eepReadWrtStrobe  when (eepReadActive = '1') else
                    eepWriteWrtStrobe when (eepWriteActive = '1') else
                    tempReadWrtStrobe when (tempReadActive = '1') else
                    ssnReadWrtStrobe  when (ssnReadActive = '1') else
                    adcReadWrtStrobe  when (adcReadActive = '1') else
                    '0';                    
   GoRead_o      <= eepReadRdStrobe  when (eepReadActive = '1') else
                    tempReadRdStrobe when (tempReadActive = '1') else
                    ssnReadRdStrobe  when (ssnReadActive = '1') else
                    adcReadRdStrobe  when (adcReadActive = '1') else
                    '0';                    
   RdWrtBytes_o  <= eepReadBytes  when (eepReadActive = '1') else
                    eepWriteBytes when (eepWriteActive = '1') else
                    tempReadBytes when (tempReadActive = '1') else
                    ssnReadBytes  when (ssnReadActive = '1') else
                    adcReadBytes  when (adcReadActive = '1') else
                    (others => '0');
   DataIn_o      <= eepReadData  when (eepReadActive = '1') else
                    eepWriteData when (eepWriteActive = '1') else
                    tempReadData when (tempReadActive = '1') else
                    ssnReadData  when (ssnReadActive = '1') else
                    adcReadData  when (adcReadActive = '1') else
                    (others => '0');
   DataInRdy_o   <= eepReadDataRdy  when (eepReadActive = '1') else
                    eepWriteDataRdy when (eepWriteActive = '1') else
                    tempReadDataRdy when (tempReadActive = '1') else
                    ssnReadDataRdy  when (ssnReadActive = '1') else
                    adcReadDataRdy  when (adcReadActive = '1') else
                    '0';
   DataOutAck_o  <= eepReadDataOutAck  when (eepReadActive = '1') else
                    tempReadDataOutAck when (tempReadActive = '1') else
                    ssnReadDataOutAck  when (ssnReadActive = '1') else
                    adcReadDataOutAck  when (adcReadActive = '1') else
                    '0';
                    
-- Register control signal mux
   eepDataRdStb_o     <= eepWriteRegRdStrobe;                
   eepDataWrtStb_o    <= eepReadRegWrtStrobe;
   eepDataRdWrtAddr_o <= eepReadRegAddr when (eepReadActive = '1') else
                         eepWriteRegAddr when (eepWriteActive = '1') else
                         "11111";                         
                    
-- Read device state machines 
   adcReadProc : process(SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1') then
            adcReadActive <= '0';
            adcReadDataOutAck <= '0';
            adcReadDataRdy <= '0';
            adcReadRdStrobe <= '0';
            adcReadWrtStrobe <= '0';
            ADC_voltage <= (others => '0');
            adcReadState <= adcReadIdle;
         else
            case adcReadState is
               when adcReadIdle =>
                  if (IfcBusy_i = '1') then
                     if (adcReadDataOutAck = '1' and DataOutRdy_i = '0') then
                        adcReadDataOutAck <= '0';
                     end if;                     
                     adcReadState <= adcReadIdle;
                  elsif (adcReadGo = '1') then
                     adcReadActive <= '1';
                     adcReadBytes <= "0000001";   -- equals 1 byte to trigger the conversion
                     if (I2C_Channel /= "00000") then  -- Only go if we are channel address is valid.
                        adcReadDataRdy <= '1';   -- Signal that we are ready to go.
                        adcReadWrtStrobe <= '1';
                        adcReadState <= adcRead_1;
                     else
                        adcReadState <= adcReadIdle;
                     end if;
                  else
                     adcReadActive <= '0';
                     adcReadDataRdy <= '0';   -- Close down shop
                     adcReadState <= adcReadIdle;
                  end if;
                  
               when adcRead_1 => -- write out the trigger first ...
                  if (adcReadWrtStrobe = '1') then
                     if (IfcBusy_i = '1') then
                        adcReadWrtStrobe <= '0';
                     end if;
                     adcReadState <= adcRead_1;
                  elsif (IfcBusy_i = '1') then
                     if (DataInAck_i = '1') then
                        adcReadState <= adcRead_2;
                     else
                        adcReadState <= adcRead_1;
                     end if;
                  else
                     adcReadState <= adcReadIdle;
                  end if;
                   
               when adcRead_2 =>
                  if (IfcBusy_i = '1') then
                     adcReadState <= adcRead_2;    -- Wait until interface goes not busy
                  else
                     adcReadDataRdy <= '0';        -- Cancel the data ready flag for the memory pointer write
                     adcReadBytes <= "0000010";    -- equals 2 bytes to read for ADC value
                     adcReadRdStrobe <= '1';       -- send the signal
                     adcReadState <= adcRead_3;    -- Now go wait for the first byte to arrive
                  end if;
                  
               when adcRead_3 => -- now initiate the read operation ... to read the previously converted data value
                  if (adcReadRdStrobe = '1') then
                     if (IfcBusy_i = '1') then
                        adcReadRdStrobe <= '0';
                     else
                        adcReadState <= adcRead_3;    -- loop until interface goes busy again
                     end if;
                  elsif (IfcBusy_i = '1') then    -- Make sure we haven't dropped out
                     if (DataOutRdy_i = '1') then  -- and capture the first byte out of the sensors                     
                        ADC_voltage(11 downto 4) <= DataOut_i(23 downto 16);
                        adcReadDataOutAck <= '1';
                        adcReadState <= adcRead_4;
                     else
                        adcReadState <= adcRead_3;
                     end if;
                  else
                     adcReadState <= adcReadIdle;
                  end if;
                  
               when adcRead_4 => -- now read the second byte
                  if (IfcBusy_i = '1') then    -- Make sure we haven't dropped out
                     if (adcReadDataOutAck = '1') then
                        if (DataOutRdy_i = '0') then
                           adcReadDataOutAck <= '0';
                        end if;
                        adcReadState <= adcRead_4;                     
                     elsif (DataOutRdy_i = '1') then  -- and capture the second byte out of the sensors
                        ADC_voltage(3 downto 0) <= DataOut_i(23 downto 20);
                        adcReadDataOutAck <= '1';
                        adcReadState <= adcReadIdle;
                     else
                        adcReadState <= adcRead_4;
                     end if;
                  else
                     adcReadState <= adcReadIdle;
                  end if;
               when others =>
                  adcReadState <= adcReadIdle;
            end case;
         end if;
      end if;
   end process;
  
   tempReadProc : process(SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1') then
            tempReadActive <= '0';
            tempReadDataOutAck <= '0';
            tempReadDataRdy <= '0';
            tempReadRdStrobe <= '0';
            tempReadWrtStrobe <= '0';
            LCB_temperature0 <= (others => '0');
            LCB_temperature1 <= (others => '0');
            PSM_temperature0 <= (others => '0');
            PSM_temperature1 <= (others => '0');
            TSM_temperature0 <= (others => '0');
            TSM_temperature1 <= (others => '0');
            AFE0_temperature0 <= (others => '0');
            AFE0_temperature1 <= (others => '0');
            AFE1_temperature0 <= (others => '0');
            AFE1_temperature1 <= (others => '0');
            tempReadState <= tempReadIdle;
         else
            case tempReadState is
               when tempReadIdle =>
                  if (IfcBusy_i = '1') then
                     if (tempReadDataOutAck = '1' and DataOutRdy_i = '0') then
                        tempReadDataOutAck <= '0';
                     end if;                     
                     tempReadState <= tempReadIdle;
                  elsif (tempReadGo = '1') then
                     tempReadActive <= '1';
                     tempReadBytes <= "0000001";   -- equals 1 address byte to set register pointer
                     if (I2C_Channel /= "00000") then  -- Only go if we are channel address is valid.
                        tempReadDataRdy <= '1';   -- Signal that we are ready to go.
                        tempReadWrtStrobe <= '1';
                     end if;
                      tempReadState <= tempRead_1;
                  else
                     tempReadActive <= '0';
                     tempReadDataRdy <= '0';   -- Close down shop
                     tempReadState <= tempReadIdle;
                  end if;
               when tempRead_1 => -- write out the address pointer first ...
                  if (tempReadWrtStrobe = '1') then
                     if (IfcBusy_i = '1') then
                        tempReadWrtStrobe <= '0';
                     end if;
                     tempReadState <= tempRead_1;
                  elsif (IfcBusy_i = '1') then
                     if (DataInAck_i = '1') then
                        tempReadState <= tempRead_2;
                     else
                        tempReadState <= tempRead_1;
                     end if;
                  else
                     tempReadState <= tempReadIdle;
                  end if;
               when tempRead_2 => -- now initiate the read operation ...
                  if (IfcBusy_i = '1') then
                     tempReadState <= tempRead_2;    -- Wait until interface goes not busy
                  else
                     tempReadDataRdy <= '0';       -- Cancel the data ready flag for the memory pointer write
                     tempReadBytes <= "0000010";   -- equals 2 bytes to read for temperature
                     tempReadRdStrobe <= '1';      -- send the signal
                     tempReadState <= tempRead_3;    -- Now go wait for the first byte to arrive
                  end if;
               when tempRead_3 => -- now initiate the read operation ...
                  if (tempReadRdStrobe = '1') then
                     if (IfcBusy_i = '1') then
                        tempReadRdStrobe <= '0';
                     else
                        tempReadState <= tempRead_3;    -- loop until interface goes busy again
                     end if;
                  elsif (IfcBusy_i = '1') then    -- Make sure we haven't dropped out
                     if (DataOutRdy_i = '1') then  -- and capture the first byte out of the sensors
                        if (I2C_DeviceAddr = TEMP0_ADDR) then
                           if (I2C_Channel(0) = '1') then
                              LCB_temperature0(11 downto 4) <= DataOut_i(7 downto 0);
                           end if;
                           if (I2C_Channel(1) = '1') then
                              PSM_temperature0(11 downto 4) <= DataOut_i(15 downto 8);
                           end if;
                           if (I2C_Channel(2) = '1') then
                              TSM_temperature0(11 downto 4) <= DataOut_i(23 downto 16);
                           end if;
                           if (I2C_Channel(3) = '1') then
                              AFE0_temperature0(11 downto 4) <= DataOut_i(31 downto 24);
                           end if;
                           if (I2C_Channel(4) = '1') then
                              AFE1_temperature0(11 downto 4) <= DataOut_i(39 downto 32);
                           end if;
                        else      
                           if (I2C_Channel(0) = '1') then
                              LCB_temperature1(11 downto 4) <= DataOut_i(7 downto 0);
                           end if;
                           if (I2C_Channel(1) = '1') then
                              PSM_temperature1(11 downto 4) <= DataOut_i(15 downto 8);
                           end if;
                           if (I2C_Channel(2) = '1') then
                              TSM_temperature1(11 downto 4) <= DataOut_i(23 downto 16);
                           end if;
                           if (I2C_Channel(3) = '1') then
                              AFE0_temperature1(11 downto 4) <= DataOut_i(31 downto 24);
                           end if;
                           if (I2C_Channel(4) = '1') then
                              AFE1_temperature1(11 downto 4) <= DataOut_i(39 downto 32);
                           end if;
                        end if;      
                        tempReadDataOutAck <= '1';
                        tempReadState <= tempRead_4;
                     else
                        tempReadState <= tempRead_3;
                     end if;
                  else
                     tempReadState <= tempReadIdle;
                  end if;
               when tempRead_4 => -- now read the second byte
                  if (IfcBusy_i = '1') then    -- Make sure we haven't dropped out
                     if (tempReadDataOutAck = '1') then
                        if (DataOutRdy_i = '0') then
                           tempReadDataOutAck <= '0';
                        end if;
                        tempReadState <= tempRead_4;                     
                     elsif (DataOutRdy_i = '1') then  -- and capture the second byte out of the sensors
                        if (I2C_DeviceAddr = TEMP0_ADDR) then
                           if (I2C_Channel(0) = '1') then
                              LCB_temperature0(3 downto 0) <= DataOut_i(7 downto 4);
                           end if;
                           if (I2C_Channel(1) = '1') then
                              PSM_temperature0(3 downto 0) <= DataOut_i(15 downto 12);
                           end if;
                           if (I2C_Channel(2) = '1') then
                              TSM_temperature0(3 downto 0) <= DataOut_i(23 downto 20);
                           end if;
                           if (I2C_Channel(3) = '1') then
                              AFE0_temperature0(3 downto 0) <= DataOut_i(31 downto 28);
                           end if;
                           if (I2C_Channel(4) = '1') then
                              AFE1_temperature0(3 downto 0) <= DataOut_i(39 downto 36);
                           end if;
                        else      
                           if (I2C_Channel(0) = '1') then
                              LCB_temperature1(3 downto 0) <= DataOut_i(7 downto 4);
                           end if;
                           if (I2C_Channel(1) = '1') then
                              PSM_temperature1(3 downto 0) <= DataOut_i(15 downto 12);
                           end if;
                           if (I2C_Channel(2) = '1') then
                              TSM_temperature1(3 downto 0) <= DataOut_i(23 downto 20);
                           end if;
                           if (I2C_Channel(3) = '1') then
                              AFE0_temperature1(3 downto 0) <= DataOut_i(31 downto 28);
                           end if;
                           if (I2C_Channel(4) = '1') then
                              AFE1_temperature1(3 downto 0) <= DataOut_i(39 downto 36);
                           end if;
                        end if;      
                        tempReadDataOutAck <= '1';
                        tempReadState <= tempReadIdle;
                     else
                        tempReadState <= tempRead_4;
                     end if;
                  else
                     tempReadState <= tempReadIdle;
                  end if;
               when others =>
                  tempReadState <= tempReadIdle;
            end case;
         end if;
      end if;
   end process;

   ssnReadProc : process(SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1') then
            ssnReadActive <= '0';
            ssnReadDataOutAck <= '0';
            ssnReadDataRdy <= '0';
            ssnReadRdStrobe <= '0';
            ssnReadWrtStrobe <= '0';
            LCB_SSN  <= (others => '0');
            PSM_SSN  <= (others => '0');
            TSM_SSN  <= (others => '0');
            AFE0_SSN <= (others => '0');
            AFE1_SSN <= (others => '0');
            ssnReadState <= ssnReadIdle;
         else
            case ssnReadState is
               when ssnReadIdle =>
                  if (IfcBusy_i = '1') then
                     if (ssnReadDataOutAck = '1' and DataOutRdy_i = '0') then
                        ssnReadDataOutAck <= '0';
                     end if;                     
                     ssnReadState <= ssnReadIdle;
                  elsif (ssnReadGo = '1') then
                     ssnReadActive <= '1';
                     ssnReadBytes <= "0000001";   -- equals 1 bytes to set register pointer
                     if (I2C_Channel /= "00000") then  -- Only go if we are channel address is valid.
                        ssnReadDataRdy <= '1';   -- Signal that we are ready to go.
                        ssnReadWrtStrobe <= '1';
                     end if;
                      ssnReadState <= ssnRead_1;
                  else
                     ssnReadActive <= '0';
                     ssnReadDataRdy <= '0';   -- Close down shop
                     ssnReadState <= ssnReadIdle;
                  end if;
               when ssnRead_1 => -- write out the address pointer first ...
                  if (ssnReadWrtStrobe = '1') then
                     if (IfcBusy_i = '1') then
                        ssnReadWrtStrobe <= '0';
                     end if;
                     ssnReadState <= ssnRead_1;
                  elsif (IfcBusy_i = '1') then
                     if (DataInAck_i = '1') then
                        ssnReadState <= ssnRead_2;
                     else
                        ssnReadState <= ssnRead_1;
                     end if;
                  else
                     ssnReadState <= ssnReadIdle;
                  end if;
               when ssnRead_2 => -- now initiate the read operation ...
                  if (IfcBusy_i = '1') then
                     ssnReadState <= ssnRead_2;    -- Wait until interface goes not busy
                  else
                     ssnReadDataRdy <= '0';       -- Cancel the data ready flag for the pointer write
                     ssnReadBytes <= "0000100";   -- equals 4 bytes to read for ssn
                     ssnReadRdStrobe <= '1';      -- send the signal
                     ssnReadState <= ssnRead_3;    -- Now go wait for the first byte to arrive
                  end if;
               when ssnRead_3 => -- now initiate the read operation ...
                  if (ssnReadRdStrobe = '1') then
                     if (IfcBusy_i = '1') then
                        ssnReadRdStrobe <= '0';
                     else
                        ssnReadState <= ssnRead_3;    -- loop until interface goes busy again
                     end if;
                  elsif (IfcBusy_i = '1') then    -- Make sure we haven't dropped out
                     if (DataOutRdy_i = '1') then  -- and capture the first byte out of the sensors
                        if (I2C_Channel(0) = '1') then
                           LCB_SSN(7 downto 0) <= DataOut_i(7 downto 0);
                        end if;
                        if (I2C_Channel(1) = '1') then
                           PSM_SSN(7 downto 0) <= DataOut_i(15 downto 8);
                        end if;
                        if (I2C_Channel(2) = '1') then
                           TSM_SSN(7 downto 0) <= DataOut_i(23 downto 16);
                        end if;
                        if (I2C_Channel(3) = '1') then
                           AFE0_SSN(7 downto 0) <= DataOut_i(31 downto 24);
                        end if;
                        if (I2C_Channel(4) = '1') then
                           AFE1_SSN(7 downto 0) <= DataOut_i(39 downto 32);
                        end if;
                        ssnReadDataOutAck <= '1';
                        ssnReadState <= ssnRead_4;
                     else
                        ssnReadState <= ssnRead_3;
                     end if;
                  else
                     ssnReadState <= ssnReadIdle;
                  end if;
               when ssnRead_4 => -- now read the second byte
                  if (IfcBusy_i = '1') then    -- Make sure we haven't dropped out
                     if (ssnReadDataOutAck = '1') then
                        if (DataOutRdy_i = '0') then
                           ssnReadDataOutAck <= '0';
                        end if;
                        ssnReadState <= ssnRead_4;                     
                     elsif (DataOutRdy_i = '1') then  -- and capture the second byte out of the sensors
                        if (I2C_Channel(0) = '1') then
                           LCB_SSN(15 downto 8) <= DataOut_i(7 downto 0);
                        end if;
                        if (I2C_Channel(1) = '1') then
                           PSM_SSN(15 downto 8) <= DataOut_i(15 downto 8);
                        end if;
                        if (I2C_Channel(2) = '1') then
                           TSM_SSN(15 downto 8) <= DataOut_i(23 downto 16);
                        end if;
                        if (I2C_Channel(3) = '1') then
                           AFE0_SSN(15 downto 8) <= DataOut_i(31 downto 24);
                        end if;
                        if (I2C_Channel(4) = '1') then
                           AFE1_SSN(15 downto 8) <= DataOut_i(39 downto 32);
                        end if;
                        ssnReadDataOutAck <= '1';
                        ssnReadState <= ssnRead_5;
                     else
                        ssnReadState <= ssnRead_4;
                     end if;
                  else
                     ssnReadState <= ssnReadIdle;
                  end if;
               when ssnRead_5 => -- now read the second byte
                  if (IfcBusy_i = '1') then    -- Make sure we haven't dropped out
                     if (ssnReadDataOutAck = '1') then
                        if (DataOutRdy_i = '0') then
                           ssnReadDataOutAck <= '0';
                        end if;
                        ssnReadState <= ssnRead_5;                     
                     elsif (DataOutRdy_i = '1') then  -- and capture the second byte out of the sensors
                        if (I2C_Channel(0) = '1') then
                           LCB_SSN(23 downto 16) <= DataOut_i(7 downto 0);
                        end if;
                        if (I2C_Channel(1) = '1') then
                           PSM_SSN(23 downto 16) <= DataOut_i(15 downto 8);
                        end if;
                        if (I2C_Channel(2) = '1') then
                           TSM_SSN(23 downto 16) <= DataOut_i(23 downto 16);
                        end if;
                        if (I2C_Channel(3) = '1') then
                           AFE0_SSN(23 downto 16) <= DataOut_i(31 downto 24);
                        end if;
                        if (I2C_Channel(4) = '1') then
                           AFE1_SSN(23 downto 16) <= DataOut_i(39 downto 32);
                        end if;
                        ssnReadDataOutAck <= '1';
                        ssnReadState <= ssnRead_6;
                     else
                        ssnReadState <= ssnRead_5;
                     end if;
                  else
                     ssnReadState <= ssnReadIdle;
                  end if;
               when ssnRead_6 => -- now read the second byte
                  if (IfcBusy_i = '1') then    -- Make sure we haven't dropped out
                     if (ssnReadDataOutAck = '1') then
                        if (DataOutRdy_i = '0') then
                           ssnReadDataOutAck <= '0';
                        end if;
                        ssnReadState <= ssnRead_6;                     
                     elsif (DataOutRdy_i = '1') then  -- and capture the second byte out of the sensors
                        if (I2C_Channel(0) = '1') then
                           LCB_SSN(31 downto 24) <= DataOut_i(7 downto 0);
                        end if;
                        if (I2C_Channel(1) = '1') then
                           PSM_SSN(31 downto 24) <= DataOut_i(15 downto 8);
                        end if;
                        if (I2C_Channel(2) = '1') then
                           TSM_SSN(31 downto 24) <= DataOut_i(23 downto 16);
                        end if;
                        if (I2C_Channel(3) = '1') then
                           AFE0_SSN(31 downto 24) <= DataOut_i(31 downto 24);
                        end if;
                        if (I2C_Channel(4) = '1') then
                           AFE1_SSN(31 downto 24) <= DataOut_i(39 downto 32);
                        end if;
                        ssnReadDataOutAck <= '1';
                        ssnReadState <= ssnReadIdle;
                     else
                        ssnReadState <= ssnRead_6;
                     end if;
                  else
                     ssnReadState <= ssnReadIdle;
                  end if;
               when others =>
                  ssnReadState <= ssnReadIdle;
            end case;
         end if;
      end if;
   end process;

   DataOutByte <= DataOut_i(7 downto 0)   when (I2C_Channel = "00001") else
                  DataOut_i(15 downto 8)  when (I2C_Channel = "00010") else
                  DataOut_i(23 downto 16) when (I2C_Channel = "00100") else
                  DataOut_i(31 downto 24) when (I2C_Channel = "01000") else
                  DataOut_i(39 downto 32) when (I2C_Channel = "10000") else
                  "00000000";
                  
   I2C_ChanRdCheck <= '1' when (I2C_Channel = "00001" or I2C_Channel = "00010" or I2C_Channel = "00100" or I2C_Channel = "01000" or I2C_Channel = "10000") else
                      '0';
                         
   eepReadProc : process(SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1') then
            eepReadActive <= '0';
            eepReadDataOutAck <= '0';
            eepReadDataRdy <= '0';
            eepReadRdStrobe <= '0';
            eepReadRegWrtStrobe <= '0';
            eepDataRegister_io <= (others => 'Z');
            eepReadState <= eepReadIdle;
         else
            case eepReadState is
               when eepReadIdle =>
                  if (eepReadGo = '1') then
                     eepReadActive <= '1';
                     eepReadBytes <= "0000010";   -- equals 2 address bytes to set eeprom address pointer
                     eepReadData <= eepMemAddr(15 downto 8); -- first out is the high address byte
                     if (I2C_ChanRdCheck = '1') then  -- Only go if the channel address is valid i.e. only one channel addressed.
                        eepReadDataRdy <= '1';
                        eepReadWrtStrobe <= '1';
                     end if;
                     eepReadState <= eepRead_1;
                  else
                     eepReadActive <= '0';
                     eepReadDataOutAck <= '0';
                     eepReadDataRdy <= '0';   -- Cancel all signs of life.
                     eepReadRegAddr <= (others => '1');
                     eepDataRegister_io <= (others => 'Z');
                     eepReadState <= eepReadIdle;
                  end if;
               when eepRead_1 => -- write out the address pointer first ...
                  if (eepReadWrtStrobe = '1') then
                     if (IfcBusy_i = '1') then
                        eepReadWrtStrobe <= '0';
                     end if;
                     eepReadState <= eepRead_1;
                  elsif (IfcBusy_i = '1') then
                     if (DataInAck_i = '1') then
                        eepReadData <= eepMemAddr(7 downto 0); -- next out is the low address byte                        
                        eepReadState <= eepRead_2;
                     else
                        eepReadState <= eepRead_1;
                     end if;
                  else
                     eepReadState <= eepReadIdle;
                  end if;
               when eepRead_2 => -- now initiate the read operation ...
                  if (IfcBusy_i = '1') then
                     eepReadState <= eepRead_2;    -- Wait until interface goes not busy
                  else
                     eepReadDataRdy <= '0';       -- Cancel the data ready flag for the memory pointer write
                     eepReadBytes <= "1000000";   -- equals 64 bytes to read
                     eepReadRdStrobe <= '1';      -- send the signal
                     eepReadState <= eepRead_3;    -- Now go wait for the first byte to arrive
                  end if;
               when eepRead_3 => -- now initiate the read operation ...
                  eepReadRegWrtStrobe <= '0'; -- cancel any write strobe to the register control 
                  if (eepReadRdStrobe = '1') then
                     if (IfcBusy_i = '1') then
                        eepReadRdStrobe <= '0';
                     else
                        eepReadState <= eepRead_3;    -- loop until interface goes busy again
                     end if;
                  elsif (IfcBusy_i = '1') then    -- Make sure we haven't dropped out
                     if (eepReadRegAddr = "01111") then  -- detect the last word to be read
                        eepReadDataOutAck <= '0';   -- Cancel the last acknowledge strobe.
                        eepReadState <= eepReadIdle;
                     else
                        eepReadRegAddr <= eepReadRegAddr + 1;
                        eepReadState <= eepRead_4;
                     end if;
                  else
                     eepReadState <= eepReadIdle;
                  end if;
               when eepRead_4 =>   -- Catch the first byte
                  if (IfcBusy_i = '1') then               
                     if (eepReadDataOutAck = '1') then
                        if (DataOutRdy_i = '0') then     -- wait for any previous data ready to go false
                           eepReadDataOutAck <= '0';
                        end if;
                        eepReadState <= eepRead_4;
                     elsif (DataOutRdy_i = '1') then
                        eepReadWord(7 downto 0) <= DataOutByte;
                        eepReadDataOutAck <= '1';
                        eepReadState <= eepRead_5;
                     else
                        eepReadState <= eepRead_4;
                     end if;
                  else
                     eepReadState <= eepReadIdle;
                  end if;
               when eepRead_5 =>   -- Catch the second byte
                  if (IfcBusy_i = '1') then
                     if (eepReadDataOutAck = '1') then
                        if (DataOutRdy_i = '0') then
                           eepReadDataOutAck <= '0';
                        end if;
                        eepReadState <= eepRead_5;
                     elsif (DataOutRdy_i = '1') then
                        eepReadWord(15 downto 8) <= DataOutByte;
                        eepReadDataOutAck <= '1';
                        eepReadState <= eepRead_6;
                     else
                        eepReadState <= eepRead_5;
                     end if;
                  else
                     eepReadState <= eepReadIdle;
                  end if;
               when eepRead_6 =>   -- Catch the third byte
                  if (IfcBusy_i = '1') then
                     if (eepReadDataOutAck = '1') then
                        if (DataOutRdy_i = '0') then
                           eepReadDataOutAck <= '0';
                        end if;
                        eepReadState <= eepRead_6;
                     elsif (DataOutRdy_i = '1') then
                        eepReadWord(23 downto 16) <= DataOutByte;
                        eepReadDataOutAck <= '1';
                        eepReadState <= eepRead_7;
                     else
                        eepReadState <= eepRead_6;
                     end if;
                  else
                     eepReadState <= eepReadIdle;
                  end if;
               when eepRead_7 =>   -- Catch the forth byte and write to the register control
                  if (IfcBusy_i = '1') then
                     if (eepReadDataOutAck = '1') then
                        if (DataOutRdy_i = '0') then
                           eepReadDataOutAck <= '0';
                        end if;
                        eepReadState <= eepRead_7;
                     elsif (DataOutRdy_i = '1') then
                        eepReadDataOutAck <= '1';
                        eepDataRegister_io <= DataOutByte & eepReadWord;
                        eepReadRegWrtStrobe <= '1'; -- write strobe to the register control 
                        eepReadState <= eepRead_3; -- now loop back and see if we have another word to catch
                     else
                        eepReadState <= eepRead_7;
                     end if;
                  else
                     eepReadState <= eepReadIdle;
                  end if;
               when others =>
                  eepReadState <= eepReadIdle;
            end case;
         end if;
      end if;
   end process;

-- Write to eeprom device state machine - This fsm writes 2 address bytes plus a page of eeprom data to the device. 
   eepWriteProc : process(SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1') then
            eepWriteActive <= '0';
            eepWriteRegRdStrobe <= '0';
            eepWriteRegAddr <= (others => '1');
            eepWriteState <= eepWriteIdle;
         else
            case eepWriteState is
               when eepWriteIdle =>
                  if (IfcBusy_i = '1') then
                     eepWriteRegRdStrobe <= '0';
                     eepWriteState <= eepWriteIdle;   -- Just loop here after coming back from a write until interface goes not busy.                      
                  elsif (eepWriteGo = '1') then    -- Detect a command to run the fsm
                     eepWriteActive <= '1';
                     eepWriteData <= eepMemAddr(15 downto 8); -- first out is the high address byte
                     if (I2C_Channel /= "00000") then  -- Only go if we are channel address is valid.
                        eepWriteDataRdy <= '1';   -- Signal that we are ready to go.
                        eepWriteWrtStrobe <= '1';
                     end if;
                     eepWriteState <= eepWrite_1;
                  else
                     eepWriteActive <= '0';
                     eepWriteDataRdy <= '0';   -- Cancel all signs of life.
                     eepWriteRegAddr <= (others => '1');
                     eepWriteState <= eepWriteIdle;
                  end if;
               when eepWrite_1 =>   -- Wait for the first data acknowledge and the interface remaining busy ...  
                  if (eepWriteWrtStrobe = '1') then
                     if (IfcBusy_i = '1') then
                        eepWriteWrtStrobe <= '0';
                     end if;
                     eepWriteState <= eepWrite_1;
                  elsif (IfcBusy_i = '1') then
                     if (DataInAck_i = '1') then
                        eepWriteData <= eepMemAddr(7 downto 0); -- next out is the low address byte                        
                        eepWriteState <= eepWrite_2;
                     else
                        eepWriteState <= eepWrite_1;
                     end if;
                  else
                     eepWriteState <= eepWriteIdle;
                  end if;
               when eepWrite_2 =>   -- loop and send 64 bytes of data to the interface
                  if (eepWriteRegAddr = "01111") then
                     eepWriteState <= eepWriteIdle;
                  else
                     eepWriteRegAddr <= eepWriteRegAddr + 1;  -- Point to the next 32-bit word in the register control module 
                     eepWriteRegRdStrobe <= '1';
                     eepWriteState <= eepWrite_3;
                  end if;
               when eepWrite_3 =>   -- capture the 32-bit word from the register control and send the first byte to the interface
                  eepWriteWord <= eepDataRegister_io;
                  if (IfcBusy_i = '1') then
                     if (DataInAck_i = '1') then
                        eepWriteData <= eepWriteWord(7 downto 0); -- low byte of word is the low addressed byte                        
                        eepWriteState <= eepWrite_4;
                     else
                        eepWriteState <= eepWrite_3;
                     end if;
                  else
                     eepWriteState <= eepWriteIdle;
                  end if;
               when eepWrite_4 =>   -- send the second byte to the interface
                  if (IfcBusy_i = '1') then
                     if (DataInAck_i = '1') then
                        eepWriteData <= eepWriteWord(15 downto 8); -- second byte of word is the next addressed byte                        
                        eepWriteState <= eepWrite_5;
                     else
                        eepWriteState <= eepWrite_4;
                     end if;
                  else
                     eepWriteState <= eepWriteIdle;
                  end if;
               when eepWrite_5 =>   -- send the third byte to the interface
                  if (IfcBusy_i = '1') then
                     if (DataInAck_i = '1') then
                        eepWriteData <= eepWriteWord(23 downto 16); -- third byte of word is the next addressed byte                        
                        eepWriteState <= eepWrite_6;
                     else
                        eepWriteState <= eepWrite_5;
                     end if;
                  else
                     eepWriteState <= eepWriteIdle;
                  end if;
               when eepWrite_6 =>   -- send the forth byte to the interface
                  if (IfcBusy_i = '1') then
                     if (DataInAck_i = '1') then
                        eepWriteData <= eepWriteWord(31 downto 24); -- forth byte of word is the next addressed byte                        
                        eepWriteState <= eepWrite_2;  -- Loop back for last word check
                     else
                        eepWriteState <= eepWrite_6;
                     end if;
                  else
                     eepWriteState <= eepWriteIdle;
                  end if;
               when others =>
                  eepWriteState <= eepWriteIdle;
            end case;
         end if;
      end if;
   end process;
   
end Behavioral;
