------------------------------------------------------------------------
-- Copyright Notice:
-- (c) Copyright 2004 - 2008, National Optical Astronomy Observatory,
-- which is operated for the National Science Foundation by the
-- Association of Universities for Research in Astronomy under
-- Cooperative Agreement No. AST-0132798.
------------------------------------------------------------------------
-- PROJECT      :   MONSOON-TORRENT
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   CFG_SERVICES
-- MODULE       :   I2C PORT MULTIPLEXOR
-- VERSION      :   2.10
------------------------------------------------------------------------
-- FILE NAME    :   I2C_Mux_V210.vhd
-- AUTHOR         :   Peter Moore
-- DATE               :   10/18/2011
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.03
------------------------------------------------------------------------
-- DESCRIPTION  :  Multiplexes the I2C telemetry and data bus to acces the various hardware modules (LCB / PSM / TSM / AFE0 / AFE1)
--                           Detects the presence of the individual I2C busses.
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
-- Buss detection:
-- Buss detection works in the following way - 
-- Step 1.  Disable the hardware drivers and enable the hardware 'receivers' - I2cSrcEn_o <= '1'; I2cSnsEn_o <= '1';
--        2.  The receivers are now active as bilateral signal translators, set the SDA SNS receivers to be outputs and toggle the SNS outputs high then low -  PsmSda_o <= '1'; wait 1us, then PsmSda_o <= '0'; etc.
--        3.  Set the SDA SNS receivers to inputs. This should leave the module side of the signal low
--        4.  Set the SDA SRC signals low, the SCL signals low and enable the hardware drivers - PsmScl_o <= '0'; PsmSda_o <= '0'; I2cSrcEn_o <= '1';
--        5.   Read the SDA SNS lines, they should always be low.
--        6.   Set the SDA SRC lines high, wait for 1us.
--        7.   Read the SDA SNS lines, Modules that are present will show a high signal, modules not present should remain low.
--        8.   Set the SCL SRC lines high to put the I2C bus signals in their quiescent state.
--   Note that the LCB bus is considered to always be present i.e. it is not tested in this way.
------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	     AUTHOR  VERSION  MODIFICATIONS
-- 06/23/2009  PCM     1.00    INITIAL VERSION
-- 10/18/2011     PCM     2.10    Cleaned up the code by making DetSdaSrcBit a  constant and forcing I2cSclEn_n_o forever low 
------------------------------------------------------------------------
-- PORT DEFINITIONS :

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity I2C_Mux_V210 is
	Port (
      SysClk_i      : in  std_logic;
      Reset_i       : in  std_logic;
-- Mux control
      ChannelAddr_i : in  std_logic_vector(4 downto 0);  -- Bit select address of module to send / receive data -- note that more than one port can be selected
      DetectMods_i  : in  std_logic;   -- pulse to command the mux to go through a bus detect function to detect connected I2C busses that are connected.
-- Status outputs
      PSM_Detect_o  : out std_logic;   -- PSM I2C bus detected
      TSM_Detect_o  : out std_logic;
      AFE0_Detect_o : out std_logic;
      AFE1_Detect_o : out std_logic;
      MuxBusy_o     : out std_logic;
-- I2C Signals
      MstrScl_i     : in  std_logic;   -- I2C Master serial data clock 
      MstrSda_i     : in  std_logic;   -- I2C Master serial data out
      MstrSda_o     : out std_logic_vector(4 downto 0);   -- I2C Master serial data in. Bits (4:0) correspond to (AFE1, AFE0, TSM, PSM, LCB).
      MstrActive_i  : in  std_logic;  -- True when an I2C transaction is in progress
      MstrWrtEn_i   : in  std_logic;  -- True when the master MstrSDA_i is valid 
-- Mux Ports
      I2cSclEn_n_o  : out std_logic;   -- Enable the I2C clock signal driver hardware
      I2cSdaEn_n_o  : out std_logic;   -- Enable the I2C data signal driver hardware
      LcbScl_o      : out std_logic;   -- LCB I2C clock signal from master
      LcbSda_io     : inout std_logic; -- LCB serial data signal
      PsmScl_o      : out std_logic;   -- Power supply clock signal
      PsmSda_io     : inout  std_logic;   -- Power supply serial data input signal, also used as output to detect I2C bus present.
      PsmSda_o      : out std_logic;   -- Power supply serial data output signal
      TsmScl_o      : out std_logic;   -- Transition module clock signal
      TsmSda_io     : inout  std_logic;
      TsmSda_o      : out std_logic;
      Afe0Scl_o     : out std_logic;   -- AFE board 0 clock signal
      Afe0Sda_io    : inout  std_logic;  
      Afe0Sda_o     : out std_logic;
      Afe1Scl_o     : out std_logic;   -- AFE board 1 clock signal
      Afe1Sda_io    : inout  std_logic;
      Afe1Sda_o     : out std_logic
   );
end I2C_Mux_V210;

architecture Behavioral of I2C_Mux_V210 is

   type DetStateType is (Det_Idle, Det_1, Det_2, Det_3, Det_4, Det_5, Det_6, Det_7);
   signal DetState : DetStateType;
   signal MstrScl, MstrSda, MstrWrtEn, MstrActive : std_logic;
   signal LcbSlct, PsmSlct, TsmSlct, Afe0Slct, Afe1Slct : std_logic;
   signal DetSclSrcBit, DetSdaSnsBit, DetectActive : std_logic;
   signal DetBootFlag : std_logic := '0';
   signal DetClk, SdaSnsDrvEnb, SigsAreQuiescent : std_logic;
   signal MuxSdaInReg : std_logic_vector(4 downto 0);
   signal DetReg      : std_logic_vector(4 downto 1);
   signal DetClkCntr  : std_logic_vector(8 downto 0);
   constant DET_TICK_VAL : std_logic_vector(8 downto 0) := "011010011"; -- = (213 - 2)  = 4us @ 53.125 Mhz SysClk.
--   constant DET_TICK_VAL : std_logic_vector(8 downto 0) := "110100111"; -- = (425 - 2)  = 4us @ 106.25 Mhz SysClk.
--   constant DET_TICK_VAL : std_logic_vector(8 downto 0) := "000000110"; -- = (8 - 2)  = for debug simulation
   constant DetSdaSrcBit : std_logic := '1';
   
begin
-- I/O assignments
   LcbScl_o   <= MstrScl when (LcbSlct = '1') else '1';
   PsmScl_o   <= MstrScl when (PsmSlct = '1') else DetSclSrcBit;
   TsmScl_o   <= MstrScl when (TsmSlct = '1') else DetSclSrcBit;
   Afe0Scl_o  <= MstrScl when (Afe0Slct = '1') else DetSclSrcBit;
   Afe1Scl_o  <= MstrScl when (Afe1Slct = '1') else DetSclSrcBit;
  
   LcbSda_io  <= MstrSda when (LcbSlct = '1' and MstrWrtEn = '1') else 'Z';
   PsmSda_o   <= MstrSda when (PsmSlct = '1' and MstrWrtEn = '1') else DetSdaSrcBit;
   TsmSda_o   <= MstrSda when (TsmSlct = '1' and MstrWrtEn = '1') else DetSdaSrcBit;
   Afe0Sda_o  <= MstrSda when (Afe0Slct = '1' and MstrWrtEn = '1') else DetSdaSrcBit;
   Afe1Sda_o  <= MstrSda when (Afe1Slct = '1' and MstrWrtEn = '1') else DetSdaSrcBit;
   
   MuxSdaInReg(0) <= LcbSda_io;
   MuxSdaInReg(1) <= PsmSda_io;
   MuxSdaInReg(2) <= TsmSda_io;
   MuxSdaInReg(3) <= Afe0Sda_io;
   MuxSdaInReg(4) <= Afe1Sda_io;
   
   PsmSda_io  <= DetSdaSnsBit when (SdaSnsDrvEnb = '1') else 'Z';
   TsmSda_io  <= DetSdaSnsBit when (SdaSnsDrvEnb = '1') else 'Z';
   Afe0Sda_io <= DetSdaSnsBit when (SdaSnsDrvEnb = '1') else 'Z';
   Afe1Sda_io <= DetSdaSnsBit when (SdaSnsDrvEnb = '1') else 'Z';
   
   PSM_Detect_o   <= DetReg(1);
   TSM_Detect_o   <= DetReg(2);
   AFE0_Detect_o  <= DetReg(3);
   AFE1_Detect_o  <= DetReg(4);
   
   MuxBusy_o <= DetectActive; 
   I2cSclEn_n_o <= '0';    -- Enable the hardware clock source driver


   MuxSigProc : process (SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (DetectActive = '1' or MstrActive_i = '0') then
            MstrScl    <= '0';
            MstrSda    <= '0';
            MstrWrtEn  <= '0';
            MstrActive <= '0';
            LcbSlct  <= '0';
            PsmSlct  <= '0';
            TsmSlct  <= '0';
            Afe0Slct <= '0';
            Afe1Slct <= '0';        
         elsif (MstrActive_i = '1') then     
            MstrScl    <= MstrScl_i;
            MstrSda    <= MstrSda_i;
            MstrWrtEn  <= MstrWrtEn_i;
            MstrActive <= MstrActive_i;
            MstrSda_o  <= MuxSdaInReg;
            LcbSlct  <= ChannelAddr_i(0);
            PsmSlct  <= ChannelAddr_i(1);
            TsmSlct  <= ChannelAddr_i(2);
            Afe0Slct <= ChannelAddr_i(3);
            Afe1Slct <= ChannelAddr_i(4);
         end if;
      end if;
   end process;
   
   DetectBusProc : process(SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (Reset_i = '1') then
            SigsAreQuiescent <= '0';
            DetSdaSnsBit <= '1';    -- to be the same as the quiescent signal
            I2cSdaEn_n_o <= '0';    -- Enable the hardware data source drivers            
            DetectActive <= '0';    -- We're not active yet
            DetSclSrcBit <= '1';    -- Keep the SCL high
--            DetSdaSrcBit <= '1';    -- Keep the SDA high
         elsif (MstrActive = '0' and SigsAreQuiescent = '1') then
            case DetState is
               when Det_Idle =>
                  if (DetBootFlag = '0' or DetectMods_i = '1') then     -- Detect the start condition
                     DetSclSrcBit <= '0';  -- Set the SCL state low so as not to disturb the hardware devices while toggling the SDA line
                     DetBootFlag  <= '1';  -- reset boot flag. 
                     DetectActive <= '1';  -- Go active.
                     DetReg <= (others => '0');  -- Clear the result register
                     DetState <= Det_1;
                 else
                     DetState <= Det_Idle;
                  end if;
                  
               when Det_1 =>
                  if (DetClk = '1') then
                     I2cSdaEn_n_o <= '1';  -- Disable the hardware source drivers. Means SDA is floating   on the I2C bus                  
                     DetSdaSnsBit <= '0';  -- set the receiver driver to low to force the SDA line
                     SdaSnsDrvEnb <= '1';  -- enable the driver to the receiver .. Switches the I2C device receivers into transmitters and sets the SDA line low at the I2C device
                     DetState <= Det_2;
                  else
                     DetState <= Det_1;
                  end if;
               when Det_2 =>     -- Delay 'till the clock sez go then set the SNS line high
                  if (DetClk = '1') then
                     DetSdaSnsBit <= '1';
                     DetState <= Det_3;
                  else
                     DetState <= Det_2;
                  end if;
               when Det_3 =>     -- Delay again and set the SNS bit low again - this assures that the translator is driving the bus end of the signal.
                  if (DetClk = '1') then
                     DetSdaSnsBit <= '0';
                     DetSclSrcBit <= '0';    -- Set the SCL state low so as not to disturb the hardware devices while toggling the SDA line
                     DetState <= Det_4;
                  else
                     DetState <= Det_3;
                  end if;
               when Det_4 =>     -- Delay again and 
                  if (DetClk = '1') then
                     SdaSnsDrvEnb <= '0'; -- Disable the SNS drivers
                     I2cSdaEn_n_o <= '0'; --Enable the hardware SDA drivers
                     DetState <= Det_5;
                  else
                     DetState <= Det_4;
                  end if;
               when Det_5 =>     -- Delay again and 
                  if (DetClk = '1') then
--                     DetSdaSrcBit <= '1'; -- set the src bit high to force the signal back into the SNS lines if the module is present (via the 800Ohm pullup resistor).
                     DetState <= Det_6;
                  else
                     DetState <= Det_5;
                  end if;
               when Det_6 =>     -- Delay again and 
                  if (DetClk = '1') then
                     DetReg <= MuxSdaInReg(4 downto 1);  -- Capture the results - The LCB is not registered here, always present
                     DetState <= Det_7;
                  else
                     DetState <= Det_6;
                  end if;
               when Det_7 =>     -- Clean up te mess
                  DetSclSrcBit <= '1';    -- Quiescent state for I2C SCL signals
                  DetectActive <= '0';    -- Go 'not busy'
                  SigsAreQuiescent <= '0';  -- reset delay flag that makes sure the I2C lines are at default
                  DetState <= Det_Idle;   -- Prepare for next time      
               when others =>
                   DetState <= Det_Idle;   -- Prepare for next time      
            end case;
         else
            DetectActive <= '0';    -- Go 'not busy'
            DetState <= Det_Idle;   -- Prepare for next time      
            DetSclSrcBit <= '1';    -- Quiescent state for all I2C signals
--            DetSdaSrcBit <= '1';
            DetSdaSnsBit <= '1';
            SigsAreQuiescent <= not MstrActive;
         end if;              
      end if;
   end process;

-- 4 us delay period for the bus detector signals
   DetClockTickProc : process(SysClk_i)
   begin
      if (rising_edge(SysClk_i)) then
         if (DetectActive = '1') then
            if (DetClkCntr >= DET_TICK_VAL) then
               DetClk <= '1';
               DetClkCntr <= (others => '0');
            else
               DetClk <= '0';
               DetClkCntr <= DetClkCntr + 1;
            end if;
         else
            DetClk <= '0';
            DetClkCntr <= (others => '0');
         end if;
      end if;
   end process;
   
end Behavioral;


