------------------------------------------------------------------------
-- Copyright Notice:                                                    
-- (c) Copyright 2004, National Optical Astronomy Observatory, which is 
-- operated for the National Science Foundation by the Association of   
-- Universities for Research in Astronomy under Cooperative Agreement   
-- No. AST-0132798.                                                     
--                                                                      
------------------------------------------------------------------------
-- PROJECT      :   MONSOON-TORRENT
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   CFG_SERVICES
-- MODULE       :   Integration REGISTER
-- VERSION      :   2.11
------------------------------------------------------------------------
-- FILE NAME    :   Int_Reg_Ver211.VHD
-- AUTHOR         :   PETER MOORE
-- DATE               :   02/27/2012
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.03
------------------------------------------------------------------------
-- DESCRIPTION  :   
-- The module is a 32 bit register with write enable 
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :

------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	       AUTHOR  VERSION  MODIFICATIONS
-- 01/26/2010   PCM     1.01     Adapted from Orange code version Seq_Mpuv465 to suit Torrent register scheme
-- 10/02/2011    PCM     2.10    Converted code to use an external One ms clock instead of generating one locally. Removed unused shutter status signal
-- 02/27/2012  PCM     2.11    Re-instated the pause command signal to pause the integration timer counter
--
------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity IntTimer_Ver211 is
   Port (SysClk_i    : in  std_logic;   -- Local clock source
         OneMsClk_i  : in  std_logic;   -- One millisecond time base
         IntTime_i   : in  std_logic_vector(31 downto 0); -- Required integration time in milliseconds	   
		   Reset_i     : in  std_logic;   -- Local reset
		   Enable_i    : in  std_logic;   -- Enable the integration counter to count up
		   Pause_i     : in  std_logic;   -- Pause the integration timer 
         TC_o        : out std_logic;   -- Terminal count (i.e. integration time) has expired
         IntCtr_o    : out std_logic_vector(31 downto 0)  -- Accumulative (current) integration time
		);
end IntTimer_Ver211;

architecture Behavioral of IntTimer_Ver211 is
   signal intcntr_run    : std_logic;
   signal IntCtr         : std_logic_vector(31 downto 0);
	signal TC             : std_logic;
   
begin

   TC   <= '1' when (IntTime_i <= IntCtr and intcntr_run = '1') else '0';
   TC_o <= TC;   
   IntCtr_o <= IntCtr;
   
------------------------------------------------------------------------
-- ACTUAL INTEGRATION TIMER COUNTER
------------------------------------------------------------------------
	process (SysClk_i)
	begin
		if (rising_edge(SysClk_i)) then
    	   	if (Reset_i = '1') then
    		    IntCtr <=  (others => '0');
			elsif (Enable_i = '1') then
				if (intcntr_run = '0') then
					IntCtr <= (others => '0');
					intcntr_run <= '1';
				elsif (OneMsClk_i = '1' and Pause_i = '0') then
					IntCtr <= IntCtr + 1;
				end if;
			else
				intcntr_run <= '0';
			end if;
		end if;
	end process;

end Behavioral;

