------------------------------------------------------------------------
-- Copyright Notice:                                                    
-- (c) Copyright 2004, National Optical Astronomy Observatory, which is 
-- operated for the National Science Foundation by the Association of   
-- Universities for Research in Astronomy under Cooperative Agreement   
-- No. AST-0132798.                                                     
--                                                                      
------------------------------------------------------------------------
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   CFG_SERVICES
-- MODULE       :   COMMAND REGISTER
-- VERSION      :   1.01
------------------------------------------------------------------------
-- FILE NAME    :  SeqCmdReg_V101.VHD
-- AUTHOR         :   PETER MOORE
-- DATE               :   01/26/2010
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.03
------------------------------------------------------------------------
-- DESCRIPTION  :   
-- The module is a 16 bit register with write enable 
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
-- CMDBITS significance.
-- BIT 0 	: ALWAYS TRUE
-- BIT 1	: EXTERNAL HARDWARE SYNC BIT
-- BIT 2	: TERMINAL COUNT FLAG FROM INTEGRATION TIMER
-- BIT 3	: START EXPOSURE FLAG; SET BY START EXPOSURE COMMAND; RESET BY EFR BIT(0) AND CONTROL REG BIT(0) = 0.
-- BIT 11:4 : START EXPOSURE VECTOR AS WRITTEN BY DHE COMMAND START EXP; RESET BY CONTROL REG BIT(0) = 0 (I.E SEQUENCER NOT RUNNING).
-- BIT 15:12: SOFT BITS WRITABLE FROM LOW ORDER BITS IN WRITE TO COMMAND REG.
------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	        AUTHOR  VERSION  MODIFICATIONS
-- 01/26/2010  PCM           1.01         Adapted from Orange Command_RegV465 ( Seq_Mpuv465) to suit Torrent register structure
--
------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity SeqCmdReg_V101 is
    Port ( 	
		   	SeqClk 		: in  std_logic;
            UserBits    : in  std_logic_vector(3 downto 0); -- appear in bits 15:12 of the CmdBits
            StartVector : in  std_logic_vector(7 downto 0); -- Start vector when SE_IN is true.
            STNC 	      : in  std_logic; -- bit 1                                        -- Sync flag input from DHE synchronization logic
           	TC 			: in  std_logic; -- bit 2                                       -- Terminal count flag from integration counter / timer logic
		   	SE_IN    	: in  std_logic; -- bit 3                                       -- Start exposure flag input - latches the StartVector value.
            SE_OUT      : out std_logic;                    -- Star exposure flag output - delayed by one clock cycle
            SE_RDY      : inout  std_logic;                 -- When MASTER false, is low true when start exposure flag not set. 
            MASTER      : in  std_logic;                    -- True when DHE is in master mode
            StartExpRst	: in  std_logic;                    -- Pulse from EFR register to clear the start exposure flag
            PAUSE       : in  std_logic;                    -- A pause has been commanded by the PAN - Used to reflect this state in the CmdBits
		   	Reset 		: in  std_logic;                    -- REset the CmdBits
            SeqRunning  : in  std_logic;                    -- Sequencer is running flag
           	CmdBits 	   : out std_logic_vector(15 downto 0) -- CmdBits out to the sequencer logic
         );
end SeqCmdReg_V101;

architecture Behavioral of SeqCmdReg_V101 is
    signal SE_Latch : std_logic;
begin

	CmdBits(0) <= '1';
	CmdBits(1) <= STNC when (SE_Latch = '1') else PAUSE;
	CmdBits(2) <= TC;
	CmdBits(3) <= SE_Latch when (MASTER = '0') else (SE_Latch and SE_RDY);
   CmdBits(15 downto 12) <= UserBits(3 downto 0);
   SE_OUT <= SE_Latch;
   SE_RDY <= SE_Latch when (SE_Latch = '0' and MASTER = '0') else 'Z';
       
	StartExpBits : process (Reset, SeqRunning, StartExpRst, SeqClk)
	begin
        if (Reset='1' or SeqRunning = '0') then
           CmdBits(11 downto 4) <=  (others => '0');
           SE_Latch <= '0';
        elsif (StartExpRst = '1') then 					-- Reset the start exposure flag from enables register
            SE_Latch  <= '0';
		elsif rising_edge (SeqClk) then
            if (SE_IN = '1') then
               CmdBits(11 downto 4) <= StartVector;	-- Load in the start vector
               SE_Latch  <= '1';							   -- Set the start exposure flag true
            end if;
		end if;
	end process;

end Behavioral;

